import json
import os
import platform
import re
import textwrap

import pytest

from conan.cli.exit_codes import ERROR_INVALID_CONFIGURATION, ERROR_GENERAL
from conan.internal.graph.graph import BINARY_INVALID
from conan.test.assets.genconanfile import GenConanfile
from conan.internal.util.files import save, load
from conan.test.utils.tools import TestClient, NO_SETTINGS_PACKAGE_ID


class TestValidate:

    def test_validate_create(self):
        client = TestClient()
        conanfile = textwrap.dedent("""
            from conan import ConanFile
            from conan.errors import ConanInvalidConfiguration
            class Pkg(ConanFile):
                settings = "os"

                def validate(self):
                    if self.info.settings.os == "Windows":
                        raise ConanInvalidConfiguration("Windows not supported")
            """)

        client.save({"conanfile.py": conanfile})

        client.run("create . --name=pkg --version=0.1 -s os=Linux")
        assert "pkg/0.1: Package '9a4eb3c8701508aa9458b1a73d0633783ecc2270' created" in client.out

        error = client.run("create . --name=pkg --version=0.1 -s os=Windows", assert_error=True)
        assert error == ERROR_INVALID_CONFIGURATION
        assert "pkg/0.1: Invalid: Windows not supported" in client.out

        client.run("graph info --require pkg/0.1 -s os=Windows")
        assert "binary: Invalid" in client.out
        assert "info_invalid: Windows not supported" in client.out

        client.run("graph info --require pkg/0.1 -s os=Windows --format json")
        myjson = json.loads(client.stdout)
        assert myjson["graph"]["nodes"]["1"]["binary"] == BINARY_INVALID
        assert myjson["graph"]["nodes"]["1"]["info_invalid"] == "Windows not supported" in client.out

    def test_validate_header_only(self):
        client = TestClient()
        conanfile = textwrap.dedent("""
            from conan import ConanFile
            from conan.errors import ConanInvalidConfiguration
            from conan.tools.build import check_min_cppstd
            class Pkg(ConanFile):
                settings = "os", "compiler"
                options = {"shared": [True, False], "header_only": [True, False],}
                default_options = {"shared": False, "header_only": True}

                def package_id(self):
                   if self.info.options.header_only:
                       self.info.clear()

                def validate(self):
                    if self.info.options.get_safe("header_only") == "False":
                        if self.info.settings.get_safe("compiler.version") == "12":
                          raise ConanInvalidConfiguration("This package cannot exist in gcc 12")
                        check_min_cppstd(self, 11)
                        # These configurations are impossible
                        if self.info.settings.os != "Windows" and self.info.options.shared:
                            raise ConanInvalidConfiguration("shared is only supported under windows")

                    # HOW CAN WE VALIDATE CPPSTD > 11 WHEN HEADER ONLY?

            """)

        client.save({"conanfile.py": conanfile})

        client.run("create . --name pkg --version=0.1 -s os=Linux -s compiler=gcc "
                   "-s compiler.version=11 -s compiler.libcxx=libstdc++11")
        assert re.search(r"Package '(.*)' created", str(client.out))

        client.run("create . --name pkg --version=0.1 -o header_only=False -s os=Linux "
                   "-s compiler=gcc -s compiler.version=12 -s compiler.libcxx=libstdc++11",
                   assert_error=True)

        assert "Invalid: This package cannot exist in gcc 12" in client.out

        client.run("create . --name pkg --version=0.1  -o header_only=False -s os=Macos "
                   "-s compiler=gcc -s compiler.version=11 -s compiler.libcxx=libstdc++11 "
                   "-s compiler.cppstd=98",
                   assert_error=True)

        assert "Invalid: Current cppstd (98) is lower than the required C++ " \
               "standard (11)" in client.out

        client.run("create . --name pkg --version=0.1  -o header_only=False -o shared=True "
                   "-s os=Macos -s compiler=gcc "
                   "-s compiler.version=11 -s compiler.libcxx=libstdc++11 -s compiler.cppstd=11",
                   assert_error=True)

        assert "Invalid: shared is only supported under windows" in client.out

    def test_validate_compatible(self):
        client = TestClient()
        conanfile = textwrap.dedent("""
            from conan import ConanFile
            from conan.errors import ConanInvalidConfiguration
            class Pkg(ConanFile):
                settings = "os"

                def validate_build(self):
                    if self.settings.os == "Windows":
                        raise ConanInvalidConfiguration("Windows not supported")

                def compatibility(self):
                    if self.settings.os == "Windows":
                        return [{"settings": [("os", "Linux")]}]
            """)

        client.save({"conanfile.py": conanfile})

        client.run("create . --name=pkg --version=0.1 -s os=Linux")
        package_id = "9a4eb3c8701508aa9458b1a73d0633783ecc2270"
        missing_id = "ebec3dc6d7f6b907b3ada0c3d3cdc83613a2b715"
        assert f"pkg/0.1: Package '{package_id}' created" in client.out

        # This is the main difference, building from source for the specified conf, fails
        client.run("create . --name=pkg --version=0.1 -s os=Windows", assert_error=True)
        assert "pkg/0.1: Cannot build for this configuration: Windows not supported" in client.out
        client.assert_listed_binary({"pkg/0.1": (missing_id, "Invalid")})

        client.run("install --requires=pkg/0.1@ -s os=Windows --build=pkg*", assert_error=True)
        assert "pkg/0.1: Cannot build for this configuration: Windows not supported" in client.out
        assert "Windows not supported" in client.out

        client.run("install --requires=pkg/0.1@ -s os=Windows")
        assert f"pkg/0.1: Main binary package '{missing_id}' missing" in client.out
        assert f"Found compatible package '{package_id}'" in client.out
        client.assert_listed_binary({"pkg/0.1": (package_id, "Cache")})

        # --build=missing means "use existing binary if possible", and compatibles are valid binaries
        client.run("install --requires=pkg/0.1@ -s os=Windows --build=missing")
        assert f"pkg/0.1: Main binary package '{missing_id}' missing" in client.out
        assert f"Found compatible package '{package_id}'" in client.out
        client.assert_listed_binary({"pkg/0.1": (package_id, "Cache")})

        client.run("graph info --requires=pkg/0.1@ -s os=Windows")
        assert f"pkg/0.1: Main binary package '{missing_id}' missing" in client.out
        assert f"Found compatible package '{package_id}'" in client.out
        assert f"package_id: {package_id}" in client.out

        client.run("graph info --requires=pkg/0.1@ -s os=Windows --build=pkg*")
        assert "binary: Invalid" in client.out

    def test_validate_remove_package_id_create(self):
        client = TestClient()
        conanfile = textwrap.dedent("""
               from conan import ConanFile
               from conan.errors import ConanInvalidConfiguration
               class Pkg(ConanFile):
                   settings = "os"

                   def validate(self):
                       if self.info.settings.os == "Windows":
                           raise ConanInvalidConfiguration("Windows not supported")

                   def package_id(self):
                       del self.info.settings.os
               """)

        client.save({"conanfile.py": conanfile})

        client.run("create . --name=pkg --version=0.1 -s os=Linux")
        assert "pkg/0.1: Package '{}' created".format(NO_SETTINGS_PACKAGE_ID) in client.out

        client.run("create . --name=pkg --version=0.1 -s os=Windows", assert_error=True)
        assert "pkg/0.1: Invalid: Windows not supported" in client.out
        client.assert_listed_binary({"pkg/0.1": ("da39a3ee5e6b4b0d3255bfef95601890afd80709",
                                                 "Invalid")})

        client.run("graph info --requires=pkg/0.1@ -s os=Windows")
        assert "package_id: {}".format(NO_SETTINGS_PACKAGE_ID) in client.out

    def test_validate_compatible_also_invalid(self):
        client = TestClient()
        conanfile = textwrap.dedent("""
           from conan import ConanFile
           from conan.errors import ConanInvalidConfiguration
           class Pkg(ConanFile):
               settings = "os", "build_type"

               def validate(self):
                   if self.info.settings.os == "Windows":
                       raise ConanInvalidConfiguration("Windows not supported")

               def compatibility(self):
                   if self.settings.build_type == "Debug" and self.settings.os != "Windows":
                       return [{"settings": [("build_type", "Release")]}]
               """)

        client.save({"conanfile.py": conanfile})

        client.run("create . --name=pkg --version=0.1 -s os=Linux -s build_type=Release")
        package_id = "c26ded3c7aa4408e7271e458d65421000e000711"
        client.assert_listed_binary({"pkg/0.1": (package_id, "Build")})
        # compatible_packges fallback works
        client.run("install --requires=pkg/0.1@ -s os=Linux -s build_type=Debug")
        client.assert_listed_binary({"pkg/0.1": (package_id, "Cache")})

        error = client.run("create . --name=pkg --version=0.1 -s os=Windows -s build_type=Release",
                           assert_error=True)

        assert error == ERROR_INVALID_CONFIGURATION
        assert "pkg/0.1: Invalid: Windows not supported" in client.out

        client.run("graph info --requires=pkg/0.1@ -s os=Windows")
        assert "binary: Invalid" in client.out

    def test_validate_compatible_also_invalid_fail(self):
        client = TestClient()
        conanfile = textwrap.dedent("""
           from conan import ConanFile
           from conan.errors import ConanInvalidConfiguration
           class Pkg(ConanFile):
               settings = "os", "build_type"

               def validate(self):
                   if self.info.settings.os == "Windows":
                       raise ConanInvalidConfiguration("Windows not supported")

               def compatibility(self):
                   if self.settings.build_type == "Debug":
                       return [{"settings": [("build_type", "Release")]}]
               """)

        client.save({"conanfile.py": conanfile})

        package_id = "c26ded3c7aa4408e7271e458d65421000e000711"
        client.run("create . --name=pkg --version=0.1 -s os=Linux -s build_type=Release")
        assert f"pkg/0.1: Package '{package_id}' created" in client.out
        # compatible_packges fallback works
        client.run("install --requires=pkg/0.1@ -s os=Linux -s build_type=Debug")
        client.assert_listed_binary({"pkg/0.1": (package_id, "Cache")})
        # Windows invalid configuration
        error = client.run("create . --name=pkg --version=0.1 -s os=Windows -s build_type=Release",
                           assert_error=True)
        assert error == ERROR_INVALID_CONFIGURATION
        assert "pkg/0.1: Invalid: Windows not supported" in client.out

        error = client.run("install --requires=pkg/0.1@ -s os=Windows -s build_type=Release",
                           assert_error=True)
        assert error == ERROR_INVALID_CONFIGURATION
        assert "pkg/0.1: Invalid: Windows not supported" in client.out

        # Windows missing binary: INVALID
        error = client.run("install --requires=pkg/0.1@ -s os=Windows -s build_type=Debug",
                           assert_error=True)
        assert error == ERROR_INVALID_CONFIGURATION
        assert "pkg/0.1: Invalid: Windows not supported" in client.out

        error = client.run("create . --name=pkg --version=0.1 -s os=Windows -s build_type=Debug",
                           assert_error=True)
        assert error == ERROR_INVALID_CONFIGURATION
        assert "pkg/0.1: Invalid: Windows not supported" in client.out

        # info
        client.run("graph info --requires=pkg/0.1@ -s os=Windows")
        assert "binary: Invalid" in client.out
        client.run("graph info --requires=pkg/0.1@ -s os=Windows -s build_type=Debug")
        assert "binary: Invalid" in client.out

    def test_validate_options(self):
        # The dependency option doesn't affect pkg package_id, so it could find a valid binary
        # in the cache. So ConanInvalidConfiguration will solve this issue.
        client = TestClient()
        client.save({"conanfile.py": GenConanfile().with_option("myoption", [1, 2, 3])
                                                   .with_default_option("myoption", 1)})
        client.run("create . --name=dep --version=0.1")
        client.run("create . --name=dep --version=0.1 -o dep/*:myoption=2")
        conanfile = textwrap.dedent("""
           from conan import ConanFile
           from conan.errors import ConanInvalidConfiguration
           class Pkg(ConanFile):
               requires = "dep/0.1"

               def validate(self):
                   if self.dependencies["dep"].options.myoption == 2:
                       raise ConanInvalidConfiguration("Option 2 of 'dep' not supported")
           """)

        client.save({"conanfile.py": conanfile})
        client.run("create . --name=pkg1 --version=0.1 -o dep/*:myoption=1")

        client.save({"conanfile.py": GenConanfile().with_requires("dep/0.1")
                                                   .with_default_option("dep/*:myoption", 2)})
        client.run("create . --name=pkg2 --version=0.1")

        client.save({"conanfile.py": GenConanfile().with_requires("pkg2/0.1", "pkg1/0.1")})
        error = client.run("install .", assert_error=True)
        assert error == ERROR_INVALID_CONFIGURATION
        assert "pkg1/0.1: Invalid: Option 2 of 'dep' not supported" in client.out

    def test_validate_requires(self):
        client = TestClient()
        client.save({"conanfile.py": GenConanfile()})
        client.run("create . --name=dep --version=0.1")
        client.run("create . --name=dep --version=0.2")
        conanfile = textwrap.dedent("""
           from conan import ConanFile
           from conan.errors import ConanInvalidConfiguration
           class Pkg(ConanFile):
               requires = "dep/0.1"

               def validate(self):
                   if self.dependencies["dep"].ref.version > "0.1":
                       raise ConanInvalidConfiguration("dep> 0.1 is not supported")
           """)

        client.save({"conanfile.py": conanfile})
        client.run("create . --name=pkg1 --version=0.1")

        client.save({"conanfile.py": GenConanfile()
                    .with_requirement("pkg1/0.1")
                    .with_requirement("dep/0.2", override=True)})
        error = client.run("install .", assert_error=True)
        assert error == ERROR_INVALID_CONFIGURATION
        assert "pkg1/0.1: Invalid: dep> 0.1 is not supported" in client.out

        client.save({"conanfile.py": GenConanfile()
                    .with_requirement("pkg1/0.1")
                    .with_requirement("dep/0.2", force=True)})
        error = client.run("install .", assert_error=True)
        assert error == ERROR_INVALID_CONFIGURATION
        assert "pkg1/0.1: Invalid: dep> 0.1 is not supported" in client.out

    def test_validate_package_id_mode(self):
        client = TestClient()
        client.save_home({"global.conf": "core.package_id:default_unknown_mode=full_package_mode"})
        conanfile = textwrap.dedent("""
          from conan import ConanFile
          from conan.errors import ConanInvalidConfiguration
          class Pkg(ConanFile):
              settings = "os"

              def validate(self):
                  if self.info.settings.os == "Windows":
                      raise ConanInvalidConfiguration("Windows not supported")
              """)
        client.save({"conanfile.py": conanfile})
        client.run("export . --name=dep --version=0.1")

        client.save({"conanfile.py": GenConanfile().with_requires("dep/0.1")})
        error = client.run("create . --name=pkg --version=0.1 -s os=Windows", assert_error=True)
        assert error == ERROR_INVALID_CONFIGURATION
        client.assert_listed_binary({"dep/0.1": ("ebec3dc6d7f6b907b3ada0c3d3cdc83613a2b715",
                                                 "Invalid")})
        client.assert_listed_binary({"pkg/0.1": ("19ad5731bb09f24646c81060bd7730d6cb5b6108",
                                                 "Build")})
        assert "ERROR: There are invalid packages:" in client.out
        assert "dep/0.1: Invalid: Windows not supported" in client.out

    def test_validate_export_pkg(self):
        # https://github.com/conan-io/conan/issues/9797
        c = TestClient()
        conanfile = textwrap.dedent("""
            from conan import ConanFile
            from conan.errors import ConanInvalidConfiguration

            class TestConan(ConanFile):
                def validate(self):
                    raise ConanInvalidConfiguration("never ever")
            """)
        c.save({"conanfile.py": conanfile})
        c.run("export-pkg . --name=test --version=1.0", assert_error=True)
        assert "ERROR: conanfile.py (test/1.0): Invalid ID: Invalid: never ever" in c.out

    def test_validate_build_export_pkg(self):
        # https://github.com/conan-io/conan/issues/9797
        c = TestClient()
        conanfile = textwrap.dedent("""
               from conan import ConanFile
               from conan.errors import ConanInvalidConfiguration

               class TestConan(ConanFile):
                   def validate_build(self):
                       raise ConanInvalidConfiguration("never ever")
               """)
        c.save({"conanfile.py": conanfile})
        c.run("export-pkg . --name=test --version=1.0", assert_error=True)
        assert "conanfile.py (test/1.0): Cannot build for this configuration: never ever" in c.out

    def test_validate_install(self):
        # https://github.com/conan-io/conan/issues/10602
        c = TestClient()
        conanfile = textwrap.dedent("""
            from conan import ConanFile
            from conan.errors import ConanInvalidConfiguration

            class TestConan(ConanFile):
                def validate(self):
                    raise ConanInvalidConfiguration("never ever")
            """)
        c.save({"conanfile.py": conanfile})
        c.run("install .", assert_error=True)
        assert "ERROR: conanfile.py: Invalid ID: Invalid: never ever" in c.out


class TestValidateCppstd:
    """ aims to be a very close to real use case of cppstd management and validation in recipes
    """
    def test_build_17_consume_14(self):
        client = TestClient()
        # simplify it a bit
        compat = textwrap.dedent("""\
            def compatibility(conanfile):
                return [{"settings": [("compiler.cppstd", v)]} for v in ("11", "14", "17", "20")]
            """)
        client.save_home({"extensions/plugins/compatibility/compatibility.py": compat})
        conanfile = textwrap.dedent("""
            from conan import ConanFile
            from conan.errors import ConanInvalidConfiguration
            class Pkg(ConanFile):
                name = "pkg"
                version = "0.1"
                settings = "compiler"

                def validate_build(self):
                    # Explicit logic instead of using check_min_cppstd that hides details
                    if int(str(self.settings.compiler.cppstd)) < 17:
                        raise ConanInvalidConfiguration("I need at least cppstd=17 to build")

                def validate(self):
                    if int(str(self.settings.compiler.cppstd)) < 14:
                        raise ConanInvalidConfiguration("I need at least cppstd=14 to be used")
            """)

        client.save({"conanfile.py": conanfile})

        settings = "-s compiler=gcc -s compiler.version=9 -s compiler.libcxx=libstdc++"
        client.run(f"create . {settings} -s compiler.cppstd=17")
        client.assert_listed_binary({"pkg/0.1": ("91faf062eb94767a31ff62a46767d3d5b41d1eff",
                                                 "Build")})
        # create with cppstd=14 fails, not enough
        client.run(f"create . {settings} -s compiler.cppstd=14", assert_error=True)
        client.assert_listed_binary({"pkg/0.1": ("36d978cbb4dc35906d0fd438732d5e17cd1e388d",
                                                 "Invalid")})
        assert "pkg/0.1: Cannot build for this configuration: I need at least cppstd=17 to build" \
               in client.out

        # Install with cppstd=14 can fallback to the previous one
        client.run(f"install --requires=pkg/0.1 {settings} -s compiler.cppstd=14")
        # 2 valid binaries, 17 and 20
        assert "pkg/0.1: Checking 2 compatible configurations" in client.out
        client.assert_listed_binary({"pkg/0.1": ("91faf062eb94767a31ff62a46767d3d5b41d1eff",
                                                 "Cache")})

        # install with not enough cppstd should fail
        client.run(f"install --requires=pkg/0.1@ {settings} -s compiler.cppstd=11",
                   assert_error=True)
        # not even trying to fallback to compatibles
        assert "pkg/0.1: Checking" not in client.out
        client.assert_listed_binary({"pkg/0.1": ("8415595b7485d90fc413c2f47298aa5fb05a5468",
                                                 "Invalid")})
        assert "I need at least cppstd=14 to be used" in client.out

    def test_header_only_14(self):
        client = TestClient()

        conanfile = textwrap.dedent("""
            from conan import ConanFile
            from conan.tools.build import check_min_cppstd
            class Pkg(ConanFile):
                name = "pkg"
                version = "0.1"
                settings = "compiler"

                def package_id(self):
                    self.info.clear()

                def validate(self):
                    check_min_cppstd(self, 14)
            """)

        client.save({"conanfile.py": conanfile})

        settings = "-s compiler=gcc -s compiler.version=9 -s compiler.libcxx=libstdc++"
        client.run(f"create . {settings} -s compiler.cppstd=17")
        client.assert_listed_binary({"pkg/0.1": ("da39a3ee5e6b4b0d3255bfef95601890afd80709",
                                                 "Build")})
        client.run(f"create . {settings} -s compiler.cppstd=14")
        client.assert_listed_binary({"pkg/0.1": ("da39a3ee5e6b4b0d3255bfef95601890afd80709",
                                                 "Build")})

        client.run(f"create . {settings} -s compiler.cppstd=11", assert_error=True)
        client.assert_listed_binary({"pkg/0.1": ("da39a3ee5e6b4b0d3255bfef95601890afd80709",
                                                 "Invalid")})
        assert "Current cppstd (11) is lower than the required C++ standard (14)" in client.out

        # Install with cppstd=14 can fallback to the previous one
        client.run(f"install --requires=pkg/0.1 {settings} -s compiler.cppstd=14")
        client.assert_listed_binary({"pkg/0.1": ("da39a3ee5e6b4b0d3255bfef95601890afd80709",
                                                 "Cache")})

        # install with not enough cppstd should fail
        client.run(f"install --requires=pkg/0.1@ {settings} -s compiler.cppstd=11",
                   assert_error=True)
        # not even trying to fallback to compatibles
        assert "pkg/0.1: Checking" not in client.out
        client.assert_listed_binary({"pkg/0.1": ("da39a3ee5e6b4b0d3255bfef95601890afd80709",
                                                 "Invalid")})
        assert "Current cppstd (11) is lower than the required C++ standard (14)" in client.out

    def test_build_17_consume_14_transitive(self):
        """ what happens if we have:
        app->engine(shared-lib)->pkg(static-lib)
        if pkg is only buildable with cppstd>=17 and needs cppstd>=14 to be consumed, but
        as it is static it becomes an implementation detail of engine, that doesn't have any
        constraint or validate() at all
        """
        client = TestClient()
        # simplify it a bit
        compat = textwrap.dedent("""\
            def compatibility(conanfile):
                return [{"settings": [("compiler.cppstd", v)]} for v in ("11", "14", "17", "20")]
            """)
        client.save_home({"extensions/plugins/compatibility/compatibility.py": compat})
        conanfile = textwrap.dedent("""
            from conan import ConanFile
            from conan.errors import ConanInvalidConfiguration
            class Pkg(ConanFile):
                name = "pkg"
                version = "0.1"
                settings = "compiler"
                package_type = "static-library"

                def validate_build(self):
                    # Explicit logic instead of using check_min_cppstd that hides details
                    if int(str(self.settings.compiler.cppstd)) < 17:
                        raise ConanInvalidConfiguration("I need at least cppstd=17 to build")

                def validate(self):
                    if int(str(self.settings.compiler.cppstd)) < 14:
                        raise ConanInvalidConfiguration("I need at least cppstd=14 to be used")
            """)
        engine = GenConanfile("engine", "0.1").with_package_type("shared-library") \
                                              .with_requires("pkg/0.1")
        app = GenConanfile("app", "0.1").with_package_type("application") \
                                        .with_requires("engine/0.1")
        client.save({"pkg/conanfile.py": conanfile,
                     "engine/conanfile.py": engine,
                     "app/conanfile.py": app})

        settings = "-s compiler=gcc -s compiler.version=9 -s compiler.libcxx=libstdc++"
        client.run(f"create pkg {settings} -s compiler.cppstd=17")
        client.assert_listed_binary({"pkg/0.1": ("91faf062eb94767a31ff62a46767d3d5b41d1eff",
                                                 "Build")})
        client.run(f"create engine {settings} -s compiler.cppstd=17")
        client.assert_listed_binary({"pkg/0.1": ("91faf062eb94767a31ff62a46767d3d5b41d1eff",
                                                 "Cache")})
        client.run(f"install app {settings} -s compiler.cppstd=17 -v")
        client.assert_listed_binary({"pkg/0.1": ("91faf062eb94767a31ff62a46767d3d5b41d1eff",
                                                 "Skip")})
        client.run(f"install app {settings} -s compiler.cppstd=14 -v")
        client.assert_listed_binary({"pkg/0.1": ("91faf062eb94767a31ff62a46767d3d5b41d1eff",
                                                 "Skip")})
        # No binary for engine exist for cppstd=11
        client.run(f"install app {settings} -s compiler.cppstd=11", assert_error=True)
        client.assert_listed_binary({"engine/0.1": ("dc24e2caf6e1fa3e8bb047ca0f5fa053c71df6db",
                                                    "Missing")})
        client.run(f"install app {settings} -s compiler.cppstd=11 --build=missing",
                   assert_error=True)
        assert 'pkg/0.1: Invalid: I need at least cppstd=14 to be used' in client.out

    def test_build_17_consume_14_transitive_erasure(self):
        """ The same as the above test:
        app->engine(shared-lib)->pkg(static-lib)
        but in this test, the engine shared-lib does "package_id()" erasure of "pkg" dependency,
        being able to reuse it then even when cppstd==11
        """
        client = TestClient()
        # simplify it a bit
        compat = textwrap.dedent("""\
            def compatibility(conanfile):
                return [{"settings": [("compiler.cppstd", v)]} for v in ("11", "14", "17", "20")]
            """)
        client.save_home({"extensions/plugins/compatibility/compatibility.py": compat})
        conanfile = textwrap.dedent("""
            from conan import ConanFile
            from conan.errors import ConanInvalidConfiguration
            class Pkg(ConanFile):
                name = "pkg"
                version = "0.1"
                settings = "compiler"
                package_type = "static-library"

                def validate_build(self):
                    # Explicit logic instead of using check_min_cppstd that hides details
                    if int(str(self.settings.compiler.cppstd)) < 17:
                        raise ConanInvalidConfiguration("I need at least cppstd=17 to build")

                def validate(self):
                    if int(str(self.settings.compiler.cppstd)) < 14:
                        raise ConanInvalidConfiguration("I need at least cppstd=14 to be used")
            """)
        engine = textwrap.dedent("""
            from conan import ConanFile
            from conan.errors import ConanInvalidConfiguration
            class Pkg(ConanFile):
                name = "engine"
                version = "0.1"
                settings = "compiler"
                package_type = "shared-library"
                requires = "pkg/0.1"

                def package_id(self):
                    del self.info.settings.compiler.cppstd
                    self.info.requires["pkg"].full_version_mode()

            """)
        app = GenConanfile("app", "0.1").with_package_type("application") \
                                        .with_requires("engine/0.1")
        client.save({"pkg/conanfile.py": conanfile,
                     "engine/conanfile.py": engine,
                     "app/conanfile.py": app})

        settings = "-s compiler=gcc -s compiler.version=9 -s compiler.libcxx=libstdc++"
        client.run(f"create pkg {settings} -s compiler.cppstd=17")
        client.assert_listed_binary({"pkg/0.1": ("91faf062eb94767a31ff62a46767d3d5b41d1eff",
                                                 "Build")})
        client.run(f"create engine {settings} -s compiler.cppstd=17")
        client.assert_listed_binary({"engine/0.1": ("493976208e9989b554704f94f9e7b8e5ba39e5ab",
                                                    "Build")})
        client.assert_listed_binary({"pkg/0.1": ("91faf062eb94767a31ff62a46767d3d5b41d1eff",
                                                 "Cache")})
        client.run(f"install app {settings} -s compiler.cppstd=17 -v")
        client.assert_listed_binary({"engine/0.1": ("493976208e9989b554704f94f9e7b8e5ba39e5ab",
                                                    "Cache")})
        client.assert_listed_binary({"pkg/0.1": ("91faf062eb94767a31ff62a46767d3d5b41d1eff",
                                                 "Skip")})
        client.run(f"install app {settings} -s compiler.cppstd=14 -v")
        client.assert_listed_binary({"engine/0.1": ("493976208e9989b554704f94f9e7b8e5ba39e5ab",
                                                    "Cache")})
        client.assert_listed_binary({"pkg/0.1": ("91faf062eb94767a31ff62a46767d3d5b41d1eff",
                                                 "Skip")})
        # No binary for engine exist for cppstd=11
        client.run(f"install app {settings} -s compiler.cppstd=11 -v")
        client.assert_listed_binary({"pkg/0.1": ("8415595b7485d90fc413c2f47298aa5fb05a5468",
                                                 "Skip")})
        client.assert_listed_binary({"engine/0.1": ("493976208e9989b554704f94f9e7b8e5ba39e5ab",
                                                    "Cache")})
        client.run(f"install app {settings} -s compiler.cppstd=11 --build=engine*",
                   assert_error=True)
        assert 'pkg/0.1: Invalid: I need at least cppstd=14 to be used' in client.out

    @pytest.mark.parametrize("use_attribute", [True, False])
    def test_exact_cppstd(self, use_attribute):
        """ Using the default cppstd_compat sometimes is not desired, and a recipe can
        explicitly opt-out this default cppstd_compat behavior, if it knows its binaries
        won't be binary compatible among them for different cppstd values
        """
        client = TestClient()
        if use_attribute:
            conanfile = textwrap.dedent("""
                from conan import ConanFile
                class Pkg(ConanFile):
                    settings = "compiler"
                    extension_properties = {"compatibility_cppstd": False}
            """)
        else:
            conanfile = textwrap.dedent("""
                from conan import ConanFile
                class Pkg(ConanFile):
                    settings = "compiler"
                    def compatibility(self):
                        self.extension_properties = {"compatibility_cppstd": False}
            """)

        client.save({"conanfile.py": conanfile})

        settings = "-s compiler=gcc -s compiler.version=9 -s compiler.libcxx=libstdc++"
        client.run(f"create . --name=pkg --version=0.1 {settings} -s compiler.cppstd=17")
        client.assert_listed_binary({"pkg/0.1": ("91faf062eb94767a31ff62a46767d3d5b41d1eff",
                                                 "Build")})

        # Install with cppstd=14 can NOT fallback to the previous one
        client.run(f"install --requires=pkg/0.1 {settings} -s compiler.cppstd=14", assert_error=True)
        assert "ERROR: Missing prebuilt package for 'pkg/0.1'" in client.out
        assert "compiler.cppstd=14" in client.out
        assert "compiler.cppstd=17" not in client.out
        client.run(f"install --requires=pkg/0.1 {settings} -s compiler.cppstd=14 --build=missing")
        client.assert_listed_binary({"pkg/0.1": ("36d978cbb4dc35906d0fd438732d5e17cd1e388d",
                                                 "Build")})
        assert "compiler.cppstd=14" in client.out
        assert "compiler.cppstd=17" not in client.out

    @pytest.mark.skipif(platform.system() == "Windows",
                        reason="Needs at least 3 cppstd values available, msvc 191 does not")
    def test_extension_properties_cppstd_compat_non_transitiveness(self):
        """
        The cppstd_compat is not transitive, so if a recipe has cppstd_compat=False,
        its dependencies will still be checked for compatibility
        """
        tc = TestClient()
        tc.save({"dep/conanfile.py": GenConanfile("dep", "1.0").with_setting("compiler"),
                 "lib/conanfile.py": GenConanfile("lib", "1.0").with_setting("compiler")
                 .with_class_attribute('extension_properties = {"compatibility_cppstd": False}')
                 .with_requirement("dep/1.0"),
                 "app/conanfile.py": GenConanfile("app", "1.0").with_setting("compiler")
                 .with_requirement("lib/1.0")})
        tc.run("create dep -s=compiler.cppstd=20")
        tc.run("create lib -s=compiler.cppstd=17")
        tc.run("create app -s=compiler.cppstd=14", assert_error=True)

        tc.run("install --requires=app/1.0 -s=compiler.cppstd=11", assert_error=True)
        assert "dep/1.0: Found compatible package" in tc.out
        assert "ERROR: Missing binary: app/1.0" in tc.out

    def test_extension_properties_make_transitive(self):
        """
        The cppstd_compat is not transitive, so if a recipe has cppstd_compat=False,
        its dependencies will still be checked for compatibility
        """
        tc = TestClient()
        tc.save({"lib/conanfile.py": GenConanfile("lib", "1.0").with_setting("compiler"),
                 "dep/conanfile.py": GenConanfile("dep", "1.0").with_setting("compiler")
                .with_class_attribute('extension_properties = {"compatibility_cppstd": False}')
                .with_requirement("lib/1.0"),
                 "conanfile.py": GenConanfile("app", "1.0").with_setting("compiler")
                .with_requirement("dep/1.0")})
        tc.run("create lib -s=compiler.cppstd=17")
        tc.run("create dep -s=compiler.cppstd=14")
        tc.run("create . -s=compiler.cppstd=14")

        compat_path = os.path.join(tc.cache_folder,
                                   "extensions/plugins/compatibility/compatibility.py")
        compat_contents = load(compat_path)

        transitive_expansions = textwrap.indent(textwrap.dedent("""
            extension_properties = getattr(conanfile, "extension_properties", {}).copy()
            for dep in conanfile.dependencies.values():
                if not dep.extension_properties.get("compatibility_cppstd", True):
                    extension_properties["compatibility_cppstd"] = False
            """), "    ")
        existing = 'extension_properties = getattr(conanfile, "extension_properties", {})'
        compat_contents = compat_contents.replace(existing, transitive_expansions)
        save(compat_path, compat_contents)

        tc.run("install --requires=app/1.0 -s=compiler.cppstd=17", assert_error=True)

        assert "Missing prebuilt package for 'app/1.0', 'dep/1.0'" in tc.out


class TestCompatibleSettingsTarget:
    """ aims to be a very close to real use case of tool being used across different settings_target
    """
    def test_settings_target_in_compatibility_method_within_recipe(self):
        client = TestClient()
        """
        test setting_target in recipe's compatibility method
        """
        tool_conanfile = textwrap.dedent("""
            from conan import ConanFile

            class Pkg(ConanFile):
                settings = "arch"
                def compatibility(self):
                    if self.settings_target.arch == "armv7":
                        return [{"settings_target": [("arch", "armv6")]}]

                def package_id(self):
                    self.info.settings_target = self.settings_target
                    for field in self.info.settings_target.fields:
                        if field != "arch":
                            self.info.settings_target.rm_safe(field)
            """)

        client.save({"conanfile.py": tool_conanfile})
        client.run("create . --name=tool --version=0.1 -s os=Linux -s:h arch=armv6 --build-require")
        package_id = client.created_package_id("tool/0.1")
        assert f"tool/0.1: Package '{package_id}' created" in client.out

        app_conanfile = textwrap.dedent("""
            from conan import ConanFile

            class Pkg(ConanFile):
                settings = "arch"
                def requirements(self):
                    self.tool_requires("tool/0.1")
            """)

        client.save({"conanfile.py": app_conanfile})
        client.run("create . --name=app --version=0.1 -s os=Linux -s:h arch=armv7")
        assert f"Found compatible package '{package_id}'" in client.out

    def test_settings_target_in_compatibility_in_global_compatibility_py(self):
        client = TestClient()
        """
        test setting_target in global compatibility method
        """
        compat = textwrap.dedent("""\
            def compatibility(self):
                if self.settings_target.arch == "armv7":
                    return [{"settings_target": [("arch", "armv6")]}]
            """)
        client.save_home({"extensions/plugins/compatibility/compatibility.py": compat})

        tool_conanfile = textwrap.dedent("""
            from conan import ConanFile

            class Pkg(ConanFile):
                settings = "arch"

                def package_id(self):
                    self.info.settings_target = self.settings_target
                    for field in self.info.settings_target.fields:
                        if field != "arch":
                            self.info.settings_target.rm_safe(field)
            """)

        client.save({"conanfile.py": tool_conanfile})
        client.run("create . --name=tool --version=0.1 -s os=Linux -s:h arch=armv6 --build-require")
        package_id = client.created_package_id("tool/0.1")

        assert f"tool/0.1: Package '{package_id}' created" in client.out

        app_conanfile = textwrap.dedent("""
            from conan import ConanFile

            class Pkg(ConanFile):
                settings = "arch"
                def requirements(self):
                    self.tool_requires("tool/0.1")
            """)

        client.save({"conanfile.py": app_conanfile})
        client.run("create . --name=app --version=0.1 -s os=Linux -s:h arch=armv7")
        assert f"Found compatible package '{package_id}'" in client.out

    def test_no_settings_target_in_recipe_but_in_compatibility_method(self):
        client = TestClient()
        """
        test settings_target in compatibility method when recipe is not a build-require
        this should not crash. When building down-stream package it should end up with
        ERROR_GENERAL instead of crash
        """

        tool_conanfile = textwrap.dedent("""
            from conan import ConanFile

            class Pkg(ConanFile):
                settings = "arch"
                def compatibility(self):
                    if self.settings_target.arch == "armv7":
                        return [{"settings_target": [("arch", "armv6")]}]
            """)

        client.save({"conanfile.py": tool_conanfile})
        client.run("create . --name=tool --version=0.1 -s os=Linux -s:h arch=armv6")
        package_id = client.created_package_id("tool/0.1")
        assert f"tool/0.1: Package '{package_id}' created" in client.out

        app_conanfile = textwrap.dedent("""
            from conan import ConanFile

            class Pkg(ConanFile):
                settings = "arch"
                def requirements(self):
                    self.tool_requires("tool/0.1")
            """)

        client.save({"conanfile.py": app_conanfile})
        error = client.run("create . --name=app --version=0.1 -s os=Linux -s:h arch=armv7",
                           assert_error=True)
        assert error == ERROR_GENERAL
        assert "ERROR: Missing prebuilt package for 'tool/0.1'" in client.out

    def test_no_settings_target_in_recipe_but_in_global_compatibility(self):
        client = TestClient()
        """
        test settings_target in global compatibility method when recipe is not a build-require
        this should not crash. When building down-stream package it should end up with
        ERROR_GENERAL instead of crash
        """
        compat = textwrap.dedent("""\
            def compatibility(self):
                if self.settings_target.arch == "armv7":
                    return [{"settings_target": [("arch", "armv6")]}]
            """)
        client.save_home({"extensions/plugins/compatibility/compatibility.py": compat})

        tool_conanfile = textwrap.dedent("""
            from conan import ConanFile

            class Pkg(ConanFile):
                settings = "arch"
            """)

        client.save({"conanfile.py": tool_conanfile})
        client.run("create . --name=tool --version=0.1 -s os=Linux -s:h arch=armv6")
        package_id = client.created_package_id("tool/0.1")
        assert f"tool/0.1: Package '{package_id}' created" in client.out

        app_conanfile = textwrap.dedent("""
            from conan import ConanFile

            class Pkg(ConanFile):
                settings = "arch"
                def requirements(self):
                    self.tool_requires("tool/0.1")
            """)

        client.save({"conanfile.py": app_conanfile})
        error = client.run("create . --name=app --version=0.1 -s os=Linux -s:h arch=armv7",
                           assert_error=True)
        assert error == ERROR_GENERAL
        assert "ERROR: Missing prebuilt package for 'tool/0.1'" in client.out

    def test_three_packages_with_and_without_settings_target(self):
        client = TestClient()
        """
        test 3 packages, tool_a and tool_b have a mutual downstream package (app), and when
        build app it should find tool_a (a compatible version of it), and find tool_b,
        and conan create should be successful.
        """
        compat = textwrap.dedent("""\
            def compatibility(self):
                if self.settings_target.arch == "armv7":
                    return [{"settings_target": [("arch", "armv6")]}]
            """)
        client.save_home({"extensions/plugins/compatibility/compatibility.py": compat})

        tool_a_conanfile = textwrap.dedent("""
            from conan import ConanFile

            class Pkg(ConanFile):
                settings = "arch"

                def package_id(self):
                    self.info.settings_target = self.settings_target
                    for field in self.info.settings_target.fields:
                        if field != "arch":
                            self.info.settings_target.rm_safe(field)
            """)

        client.save({"conanfile.py": tool_a_conanfile})
        client.run("create . --name=tool_a --version=0.1 "
                   "-s os=Linux -s:h arch=armv6 -s:b arch=x86_64 --build-require")
        package_id_tool_a = client.created_package_id("tool_a/0.1")
        assert f"tool_a/0.1: Package '{package_id_tool_a}' created" in client.out

        tool_b_conanfile = textwrap.dedent("""
            from conan import ConanFile

            class Pkg(ConanFile):
                settings = "arch"
            """)

        client.save({"conanfile.py": tool_b_conanfile})
        client.run("create . --name=tool_b --version=0.1 -s os=Linux "
                   "-s arch=x86_64 -s:b arch=x86_64")
        package_id_tool_b = client.created_package_id("tool_b/0.1")
        assert f"tool_b/0.1: Package '{package_id_tool_b}' created" in client.out

        app_conanfile = textwrap.dedent("""
            from conan import ConanFile

            class Pkg(ConanFile):
                settings = "arch"
                def requirements(self):
                    self.tool_requires("tool_a/0.1")
                    self.tool_requires("tool_b/0.1")
            """)

        client.save({"conanfile.py": app_conanfile})
        client.run("create . --name=app --version=0.1 -s os=Linux -s:h arch=armv7 -s:b arch=x86_64")
        assert f"Found compatible package '{package_id_tool_a}'" in client.out
        assert package_id_tool_b in client.out

    def test_settings_target_in_compatibility_method_within_recipe_package_info(self):
        # https://github.com/conan-io/conan/issues/14823
        client = TestClient()
        tool_conanfile = textwrap.dedent("""
            from conan import ConanFile

            class Pkg(ConanFile):
                settings = "arch"

                def compatibility(self):
                    return [{'settings': [('arch', 'armv6')]}]

                def package_info(self):
                    # This used to crash
                    self.settings_target.get_safe('compiler.link_time_optimization')
            """)

        client.save({"conanfile.py": tool_conanfile})
        client.run("create . --name=tool --version=0.1 -s os=Linux -s:b arch=armv6 --build-require")
        package_id = client.created_package_id("tool/0.1")
        assert f"tool/0.1: Package '{package_id}' created" in client.out

        client.run("install --tool-requires=tool/0.1 -s os=Linux -s:b arch=armv7")
        # This used to crash, not anymore
        assert f"Found compatible package '{package_id}'" in client.out



import json
import os
import textwrap
import uuid
from unittest import mock

import bottle
import pytest
from bottle import static_file, HTTPError, request
from webtest import TestApp

from conan.internal.errors import NotFoundException
from conan.errors import ConanException
from conan.test.utils.file_server import TestFileServer
from conan.test.utils.test_files import temp_folder
from conan.test.utils.tools import TestClient
from conan.internal.util.files import save, load, rmdir, mkdir, remove


class TestDownloadCacheBackupSources:
    def test_users_download_cache_summary(self):
        def custom_download(this, url, filepath, *args, **kwargs):  # noqa
            assert not url.startswith("http://myback")
            save(filepath, f"Hello, world!")

        with mock.patch("conan.internal.rest.file_downloader.FileDownloader.download",
                        custom_download):
            client = TestClient(default_server_user=True, light=True)
            tmp_folder = temp_folder()
            client.save_home(
                {"global.conf": f"core.sources:download_cache={tmp_folder}\n"
                                "core.sources:download_urls=['origin', 'http://myback']"})
            sha256 = "315f5bdb76d078c43b8ac0064e4a0164612b1fce77c869345bfc94c75894edd3"
            conanfile = textwrap.dedent(f"""
                from conan import ConanFile
                from conan.tools.files import download
                class Pkg(ConanFile):
                   def source(self):
                       download(self, "http://localhost:5000/myfile.txt", "myfile.txt",
                                sha256="{sha256}")
                """)
            client.save({"conanfile.py": conanfile})
            client.run("source .")

            assert 2 == len(os.listdir(os.path.join(tmp_folder, "s")))
            content = json.loads(load(os.path.join(tmp_folder, "s", sha256 + ".json")))
            assert "http://localhost:5000/myfile.txt" in content["references"]["unknown"]
            assert len(content["references"]["unknown"]) == 1

            conanfile = textwrap.dedent(f"""
                from conan import ConanFile
                from conan.tools.files import download
                class Pkg2(ConanFile):
                    name = "pkg"
                    version = "1.0"
                    def source(self):
                        download(self, "http://localhost.mirror:5000/myfile.txt", "myfile.txt",
                                 sha256="{sha256}")
                """)
            client.save({"conanfile.py": conanfile})
            client.run("source .")

            assert 2 == len(os.listdir(os.path.join(tmp_folder, "s")))
            content = json.loads(load(os.path.join(tmp_folder, "s", sha256 + ".json")))
            assert "http://localhost.mirror:5000/myfile.txt" in content["references"]["pkg/1.0"]
            assert len(content["references"]["pkg/1.0"]) == 1

            # Ensure the cache is working and we didn't break anything by modifying the summary
            client.run("source .")
            assert "Downloading file" not in client.out

            client.run("create . --user=barbarian")
            content = json.loads(load(os.path.join(tmp_folder, "s", sha256 + ".json")))
            assert content["references"]["pkg/1.0@barbarian"] == \
                   ["http://localhost.mirror:5000/myfile.txt"]

            client.run("create . --user=barbarian --channel=stable")
            content = json.loads(load(os.path.join(tmp_folder, "s", sha256 + ".json")))
            assert content["references"]["pkg/1.0@barbarian/stable"] == \
                   ["http://localhost.mirror:5000/myfile.txt"]

    @pytest.fixture(autouse=True)
    def _setup(self):
        self.client = TestClient(default_server_user=True, light=True)
        self.file_server = TestFileServer()
        self.client.servers["file_server"] = self.file_server
        self.download_cache_folder = temp_folder()

    def test_upload_sources_backup(self):
        http_server_base_folder_backups = os.path.join(self.file_server.store, "backups")
        http_server_base_folder_internet = os.path.join(self.file_server.store, "internet")

        save(os.path.join(http_server_base_folder_internet, "myfile.txt"), "Hello, world!")
        save(os.path.join(self.file_server.store, "mycompanystorage", "mycompanyfile.txt"),
             "Business stuff")
        save(os.path.join(http_server_base_folder_internet, "duplicated1.txt"), "I am duplicated #1")
        save(os.path.join(self.file_server.store, "mycompanystorage2", "duplicated2.txt"),
             "I am duplicated #2")

        hello_world_sha256 = "315f5bdb76d078c43b8ac0064e4a0164612b1fce77c869345bfc94c75894edd3"
        mycompanyfile_sha256 = "7f1d5d6ae44eb93061b0e07661bd8cbac95a7c51fa204570bf7e24d877d4a224"
        duplicated1_sha256 = "744aca0436e2e355c285fa926a37997df488544589cec2260fc9db969a3c78df"
        duplicated2_sha256 = "66ba2ba05211da3b0b0cb0a08f18e1d9077e7321c2be27887371ec37fade376d"

        conanfile = textwrap.dedent(f"""
            from conan import ConanFile
            from conan.tools.files import download
            class Pkg2(ConanFile):
                name = "pkg"
                version = "1.0"
                def source(self):
                    download(self, "{self.file_server.fake_url}/internet/myfile.txt",
                            "myfile.txt",
                             sha256="{hello_world_sha256}")

                    download(self, "{self.file_server.fake_url}/mycompanystorage/mycompanyfile.txt",
                            "mycompanyfile.txt",
                             sha256="{mycompanyfile_sha256}")

                    download(self, ["{self.file_server.fake_url}/mycompanystorage/duplicated1.txt",
                                    "{self.file_server.fake_url}/internet/duplicated1.txt"],
                            "duplicated1.txt",
                            sha256="{duplicated1_sha256}")

                    download(self, ["{self.file_server.fake_url}/internet/duplicated1.txt"],
                            "duplicated1.txt",
                            sha256="{duplicated1_sha256}")

                    download(self, ["{self.file_server.fake_url}/mycompanystorage/duplicated2.txt",
                                    "{self.file_server.fake_url}/mycompanystorage2/duplicated2.txt"],
                            "duplicated2.txt",
                            sha256="{duplicated2_sha256}")

                    download(self, "{self.file_server.fake_url}/mycompanystorage2/duplicated2.txt",
                            "duplicated2.txt",
                            sha256="{duplicated2_sha256}")
            """)

        # Ensure that a None url is only warned about but no exception is thrown,
        # this is possible in CI systems that substitute an env var and fail to give it a value
        self.client.save_home(
            {"global.conf": f"core.sources:download_cache={self.download_cache_folder}\n"
                            f"core.sources:download_urls=[None, 'origin', '{self.file_server.fake_url}/backups/']\n"
                            f"core.sources:upload_url={self.file_server.fake_url}/backups/"})

        self.client.save({"conanfile.py": conanfile})
        self.client.run("create .", assert_error=True)
        assert "Trying to download sources from None backup remote" in self.client.out

        self.client.save_home(
            {"global.conf": f"core.sources:download_cache={self.download_cache_folder}\n"
                            f"core.sources:download_urls=['origin', '{self.file_server.fake_url}/backups/']\n"
                            f"core.sources:upload_url={self.file_server.fake_url}/backups/\n"
                            f"core.sources:exclude_urls=['{self.file_server.fake_url}/mycompanystorage/', "
                            f"'{self.file_server.fake_url}/mycompanystorage2/']"})
        self.client.run("create .")

        self.client.run("upload * -c -r=default")

        # To test #15501 that it works even with extra files in the local cache
        extra_file_path = os.path.join(self.download_cache_folder, "s", "extrafile")
        save(extra_file_path, "Hello, world!")
        self.client.run("upload * -c -r=default", assert_error=True)
        assert "Missing metadata file for backup source" in self.client.out
        os.unlink(extra_file_path)

        server_contents = os.listdir(http_server_base_folder_backups)
        assert hello_world_sha256 in server_contents
        assert hello_world_sha256 + ".json" in server_contents
        # Its only url is excluded, should not be there
        assert mycompanyfile_sha256 not in server_contents
        assert mycompanyfile_sha256 + ".json" not in server_contents
        # It has an excluded url, but another that is not, it should be there
        assert duplicated1_sha256 in server_contents
        assert duplicated1_sha256 + ".json" in server_contents
        # All its urls are excluded, it shoud not be there
        assert duplicated2_sha256 not in server_contents
        assert duplicated2_sha256 + ".json" not in server_contents

        self.client.run("upload * -c -r=default")
        assert "already in server, skipping upload" in self.client.out

        # Clear de local cache
        rmdir(self.download_cache_folder)
        mkdir(self.download_cache_folder)
        # Everything the same, but try to download from backup first
        self.client.save_home(
            {"global.conf": f"core.sources:download_cache={self.download_cache_folder}\n"
                            f"core.sources:download_urls=['{self.file_server.fake_url}/backups/', 'origin']\n"
                            f"core.sources:upload_url={self.file_server.fake_url}/backups/\n"
                            f"core.sources:exclude_urls=['{self.file_server.fake_url}/mycompanystorage/', '{self.file_server.fake_url}/mycompanystorage2/']"})
        self.client.run("source .")
        assert f"Sources for {self.file_server.fake_url}/internet/myfile.txt found in remote backup" in self.client.out
        assert f"File {self.file_server.fake_url}/mycompanystorage/mycompanyfile.txt not found in {self.file_server.fake_url}/backups/" in self.client.out

        # Ensure defaults backup folder works if it's not set in global.conf
        # (The rest is needed to exercise the rest of the code)
        self.client.save_home(
            {"global.conf": f"core.sources:download_urls=['{self.file_server.fake_url}/backups/', 'origin']\n"
                            f"core.sources:exclude_urls=['{self.file_server.fake_url}/mycompanystorage/', '{self.file_server.fake_url}/mycompanystorage2/']"})
        self.client.run("source .")
        assert f"Sources for {self.file_server.fake_url}/internet/myfile.txt found in remote backup" in self.client.out
        assert f"File {self.file_server.fake_url}/mycompanystorage/mycompanyfile.txt not found in {self.file_server.fake_url}/backups/" in self.client.out

    def test_unknown_handling(self):
        http_server_base_folder_internet = os.path.join(self.file_server.store, "internet")

        save(os.path.join(http_server_base_folder_internet, "myfile.txt"), "Hello, world!")

        sha256 = "315f5bdb76d078c43b8ac0064e4a0164612b1fce77c869345bfc94c75894edd3"
        conanfile = textwrap.dedent(f"""
                            from conan import ConanFile
                            from conan.tools.files import download
                            class Pkg2(ConanFile):
                                name = "pkg"
                                def source(self):
                                    download(self, "{self.file_server.fake_url}/internet/myfile.txt",
                                             "myfile.txt",
                                             sha256="{sha256}")
                            """)

        self.client.save_home(
            {"global.conf": f"core.sources:download_cache={self.download_cache_folder}\n"
                            f"core.sources:download_urls=['origin']\n"
                            f"core.sources:upload_url={self.file_server.fake_url}/backups/"})

        self.client.save({"conanfile.py": conanfile})
        self.client.run("source .")
        json_path = os.path.join(self.download_cache_folder, "s", sha256 + ".json")
        contents = json.loads(load(json_path))
        assert "unknown" in contents["references"]

        s_folder = os.path.join(self.download_cache_folder, "s")
        assert len(os.listdir(s_folder)) == 2

        self.client.run("export . --version=1.0")
        self.client.run("upload * -c -r=default")
        assert "No backup sources files to upload" in self.client.out

        self.client.run("cache clean --backup-sources")
        assert len(os.listdir(s_folder)) == 0

    def test_download_origin_first(self):
        http_server_base_folder_internet = os.path.join(self.file_server.store, "internet")

        save(os.path.join(http_server_base_folder_internet, "myfile.txt"), "Hello, world!")

        sha256 = "315f5bdb76d078c43b8ac0064e4a0164612b1fce77c869345bfc94c75894edd3"
        conanfile = textwrap.dedent(f"""
                    from conan import ConanFile
                    from conan.tools.files import download
                    class Pkg2(ConanFile):
                        name = "pkg"
                        version = "1.0"
                        def source(self):
                            download(self, "{self.file_server.fake_url}/internet/myfile.txt", "myfile.txt",
                                     sha256="{sha256}")
                    """)

        self.client.save_home(
            {"global.conf": f"core.sources:download_cache={self.download_cache_folder}\n"
                            f"core.sources:download_urls=['origin', '{self.file_server.fake_url}/backup/']\n"
                            f"core.sources:upload_url={self.file_server.fake_url}/backup/"})

        self.client.save({"conanfile.py": conanfile})
        self.client.run("create .")
        self.client.run("upload * -c -r=default")

        rmdir(self.download_cache_folder)

        self.client.run("source .")
        assert f"Sources for {self.file_server.fake_url}/internet/myfile.txt found in origin" in self.client.out
        self.client.run("source .")
        assert f"Source {self.file_server.fake_url}/internet/myfile.txt retrieved from local download cache" in self.client.out

    def test_download_origin_last(self):
        http_server_base_folder_internet = os.path.join(self.file_server.store, "internet")

        save(os.path.join(http_server_base_folder_internet, "myfile.txt"), "Hello, world!")

        sha256 = "315f5bdb76d078c43b8ac0064e4a0164612b1fce77c869345bfc94c75894edd3"
        conanfile = textwrap.dedent(f"""
                    from conan import ConanFile
                    from conan.tools.files import download
                    class Pkg2(ConanFile):
                        name = "pkg"
                        version = "1.0"
                        def source(self):
                            download(self, "{self.file_server.fake_url}/internet/myfile.txt", "myfile.txt",
                                     sha256="{sha256}")
                    """)

        self.client.save_home(
            {"global.conf": f"core.sources:download_cache={self.download_cache_folder}\n"
                            f"core.sources:download_urls=['{self.file_server.fake_url}/backup', 'origin']\n"
                            f"core.sources:upload_url={self.file_server.fake_url}/backup"})

        self.client.save({"conanfile.py": conanfile})
        self.client.run("create . -vv")
        assert f"WARN: File {self.file_server.fake_url}/internet/myfile.txt not found in {self.file_server.fake_url}/backup/" in self.client.out
        assert f"Downloaded {self.file_server.fake_url}/internet/myfile.txt from {self.file_server.fake_url}/internet/myfile.txt"
        self.client.run("upload * -c -r=default")

        rmdir(self.download_cache_folder)

        self.client.run("source .")
        assert f"Sources for {self.file_server.fake_url}/internet/myfile.txt found in remote backup" in self.client.out
        self.client.run("source .")
        assert f"Source {self.file_server.fake_url}/internet/myfile.txt retrieved from local download cache" in self.client.out

    def test_sources_backup_server_error_500(self):
        conanfile = textwrap.dedent(f"""
           from conan import ConanFile
           from conan.tools.files import download
           class Pkg2(ConanFile):
               name = "pkg"
               version = "1.0"
               def source(self):
                   download(self, "{self.file_server.fake_url}/internet/myfile.txt", "myfile.txt",
                            sha256="unused")
           """)

        self.client.save_home(
            {"global.conf": f"core.sources:download_cache={self.download_cache_folder}\n"
                            f"core.sources:download_urls=['{self.file_server.fake_url}/internal_error/', "
                            f"'{self.file_server.fake_url}/unused/']\n"})
        self.client.save({"conanfile.py": conanfile})
        self.client.run("create .", assert_error=True)
        assert "ConanException: Error 500 downloading file" in self.client.out

    def test_sources_backup_server_not_found(self):
        conanfile = textwrap.dedent(f"""
           from conan import ConanFile
           from conan.tools.files import download
           class Pkg2(ConanFile):
               name = "pkg"
               version = "1.0"
               def source(self):
                   download(self, "{self.file_server.fake_url}/internet/myfile.txt", "myfile.txt",
                            sha256="unused")
           """)

        self.client.save_home(
            {"global.conf": f"core.sources:download_cache={self.download_cache_folder}\n"
                            f"core.download:retry=0\n"
                            f"core.sources:download_urls=['error_server', 'origin']\n"})

        self.client.save({"conanfile.py": conanfile})
        self.client.run("create .", assert_error=True)
        assert "ConanException: Error downloading file" in self.client.out

    def test_upload_sources_backup_creds_needed(self):
        client = TestClient(default_server_user=True, light=True)
        download_cache_folder = temp_folder()
        http_server_base_folder_backups = temp_folder()
        http_server_base_folder_internet = temp_folder()

        save(os.path.join(http_server_base_folder_internet, "myfile.txt"), "Hello, world!")

        class TestFileServer2:

            def __init__(self):
                self.fake_url = "http://fake%s.com" % str(uuid.uuid4()).replace("-", "")
                self.root_app = bottle.Bottle()
                self.app = TestApp(self.root_app)
                self._attach_to(self.root_app)

            @staticmethod
            def _attach_to(app):
                def valid_auth(token):
                    auth = request.headers.get("Authorization")
                    if auth == f"Bearer {token}":
                        return
                    return HTTPError(401, "Authentication required")

                @app.route("/internet/<file>", method=["GET"])
                def get_internet_file(file):
                    return static_file(file, http_server_base_folder_internet)

                @app.route("/downloader/<file>", method=["GET"])
                def get_file(file):
                    ret = valid_auth("mytoken")
                    return ret or static_file(file, http_server_base_folder_backups)

                @app.route("/uploader/<file>", method=["PUT"])
                def put_file(file):
                    ret = valid_auth("myuploadtoken")
                    if ret:
                        return ret
                    dest = os.path.join(http_server_base_folder_backups, file)
                    with open(dest, 'wb') as f:
                        f.write(request.body.read())

        http_server = TestFileServer2()
        client.servers["file_server"] = http_server

        sha256 = "315f5bdb76d078c43b8ac0064e4a0164612b1fce77c869345bfc94c75894edd3"
        conanfile = textwrap.dedent(f"""
            from conan import ConanFile
            from conan.tools.files import download, load
            class Pkg2(ConanFile):
                name = "pkg"
                version = "1.0"
                def source(self):
                    download(self, "{http_server.fake_url}/internet/myfile.txt", "myfile.txt",
                             sha256="{sha256}")
                    self.output.info(f"CONTENT: {{load(self, 'myfile.txt')}}")
            """)

        client.save_home(
            {"global.conf": f"core.sources:download_cache={download_cache_folder}\n"
                            f"core.sources:download_urls=['{http_server.fake_url}/downloader/', 'origin']\n"
                            f"core.sources:upload_url={http_server.fake_url}/uploader/"})

        client.save({"conanfile.py": conanfile})
        client.run("create .", assert_error=True)
        assert f"ConanException: The source backup server '{http_server.fake_url}" \
               f"/downloader/' needs authentication" in client.out
        content = {"credentials": [
            {"url": f"{http_server.fake_url}", "token": "mytoken"}
        ]}
        save(os.path.join(client.cache_folder, "source_credentials.json"), json.dumps(content))

        client.run("create .")
        assert "CONTENT: Hello, world!" in client.out
        client.run("upload * -c -r=default", assert_error=True)
        assert f"The source backup server '{http_server.fake_url}" \
               f"/uploader/' needs authentication" in client.out
        content = {"credentials": [
            {"url": f"{http_server.fake_url}", "token": "myuploadtoken"}
        ]}
        # Now use the correct UPLOAD token
        save(os.path.join(client.cache_folder, "source_credentials.json"), json.dumps(content))
        client.run("upload * -c -r=default --force")  # need --force to guarantee cached updated

        server_contents = os.listdir(http_server_base_folder_backups)
        assert sha256 in server_contents
        assert sha256 + ".json" in server_contents

        client.run("upload * -c -r=default")
        assert "already in server, skipping upload" in client.out

        content = {"credentials": [
            {"url": f"{http_server.fake_url}", "token": "mytoken"}
        ]}

        save(os.path.join(client.cache_folder, "source_credentials.json"), json.dumps(content))
        rmdir(download_cache_folder)

        # Remove the "remote" myfile.txt so if it raises
        # we know it tried to download the original source
        os.remove(os.path.join(http_server_base_folder_internet, "myfile.txt"))

        client.run("source .")
        assert f"Sources for {http_server.fake_url}/internet/myfile.txt found in remote backup" in client.out
        assert "CONTENT: Hello, world!" in client.out
        client.run("source .")
        assert f"Source {http_server.fake_url}/internet/myfile.txt retrieved from local download cache" in client.out
        assert "CONTENT: Hello, world!" in client.out

    def test_download_sources_multiurl(self):
        http_server_base_folder_internet = os.path.join(self.file_server.store, "internet")
        http_server_base_folder_backup = os.path.join(self.file_server.store, "backup")
        http_server_base_folder_downloader = os.path.join(self.file_server.store, "downloader")

        save(os.path.join(http_server_base_folder_internet, "myfile.txt"), "Hello, world!")
        sha256 = "315f5bdb76d078c43b8ac0064e4a0164612b1fce77c869345bfc94c75894edd3"

        conanfile = textwrap.dedent(f"""
            from conan import ConanFile
            from conan.tools.files import download
            class Pkg2(ConanFile):
                name = "pkg"
                version = "1.0"
                def source(self):
                    download(self, "{self.file_server.fake_url}/internet/myfile.txt", "myfile.txt",
                             sha256="{sha256}")
            """)

        self.client.save_home(
            {"global.conf": f"core.sources:download_cache={self.download_cache_folder}\n"
                            f"core.sources:upload_url={self.file_server.fake_url}/backup/\n"
                            f"core.sources:download_urls=['origin', '{self.file_server.fake_url}/downloader/', "
                            f"'{self.file_server.fake_url}/backup/']"})

        self.client.save({"conanfile.py": conanfile})
        self.client.run("create .")
        self.client.run("upload * -c -r=default")
        # We upload files to second backup,
        # to ensure that the first one gets skipped in the list but finds in the second one
        server_contents = os.listdir(http_server_base_folder_backup)
        assert sha256 in server_contents
        assert sha256 + ".json" in server_contents

        rmdir(self.download_cache_folder)
        # Remove the "remote" myfile.txt so if it raises
        # we know it tried to download the original source
        remove(os.path.join(http_server_base_folder_internet, "myfile.txt"))

        self.client.run("source .")
        assert f"Sources for {self.file_server.fake_url}/internet/myfile.txt found in remote backup {self.file_server.fake_url}/backup/" in self.client.out

        # And if the first one has them, prefer it before others in the list
        save(os.path.join(http_server_base_folder_downloader, sha256),
             load(os.path.join(http_server_base_folder_backup, sha256)))
        save(os.path.join(http_server_base_folder_downloader, sha256 + ".json"),
             load(os.path.join(http_server_base_folder_backup, sha256 + ".json")))
        rmdir(self.download_cache_folder)
        self.client.run("source .")
        assert f"Sources for {self.file_server.fake_url}/internet/myfile.txt found in remote backup {self.file_server.fake_url}/downloader/" in self.client.out

    def test_list_urls_miss(self):
        def custom_download(this, url, *args, **kwargs):  # noqa
            raise NotFoundException()

        with mock.patch("conan.internal.rest.file_downloader.FileDownloader.download",
                        custom_download):
            client = TestClient(default_server_user=True, light=True)
            download_cache_folder = temp_folder()

            sha256 = "315f5bdb76d078c43b8ac0064e4a0164612b1fce77c869345bfc94c75894edd3"

            conanfile = textwrap.dedent(f"""
                from conan import ConanFile
                from conan.tools.files import download
                class Pkg2(ConanFile):
                    name = "pkg"
                    version = "1.0"
                    def source(self):
                        download(self, "http://fake/myfile.txt", "myfile.txt", sha256="{sha256}")
                """)

            client.save_home(
                {"global.conf": f"core.sources:download_cache={download_cache_folder}\n"
                                f"core.sources:download_urls=['origin', 'http://extrafake/']\n"})
            client.save({"conanfile.py": conanfile})
            client.run("source .", assert_error=True)
            assert "WARN: Sources for http://fake/myfile.txt failed in 'origin'" in client.out
            assert "WARN: Checking backups" in client.out
            assert "NotFoundException: File http://fake/myfile.txt " \
                   "not found in ['origin', 'http://extrafake/']" in client.out

    def test_ok_when_origin_breaks_midway_list(self):
        http_server_base_folder_backup2 = os.path.join(self.file_server.store, "backup2")

        sha256 = "315f5bdb76d078c43b8ac0064e4a0164612b1fce77c869345bfc94c75894edd3"
        save(os.path.join(http_server_base_folder_backup2, sha256), "Hello, world!")
        save(os.path.join(http_server_base_folder_backup2, sha256 + ".json"), '{"references": {}}')

        conanfile = textwrap.dedent(f"""
           from conan import ConanFile
           from conan.tools.files import download
           class Pkg2(ConanFile):
               name = "pkg"
               version = "1.0"
               def source(self):
                   download(self, "{self.file_server.fake_url}/internal_error/myfile.txt", "myfile.txt",
                            sha256="{sha256}")
           """)

        self.client.save_home(
            {"global.conf": "tools.files.download:retry=0\n"
                            f"core.sources:download_cache={self.download_cache_folder}\n"
                            f"core.sources:download_urls=['{self.file_server.fake_url}/empty/', "
                            f"'origin', '{self.file_server.fake_url}/backup2/']\n"})

        self.client.save({"conanfile.py": conanfile})
        self.client.run("create .")
        assert f"Sources for {self.file_server.fake_url}/internal_error/myfile.txt found in remote backup {self.file_server.fake_url}/backup2/" in self.client.out

    def test_ok_when_origin_authorization_error(self):
        client = TestClient(default_server_user=True, light=True)
        download_cache_folder = temp_folder()
        http_server_base_folder_backup1 = temp_folder()
        http_server_base_folder_backup2 = temp_folder()

        sha256 = "315f5bdb76d078c43b8ac0064e4a0164612b1fce77c869345bfc94c75894edd3"
        save(os.path.join(http_server_base_folder_backup2, sha256), "Hello, world!")
        save(os.path.join(http_server_base_folder_backup2, sha256 + ".json"), '{"references": {}}')

        class TestFileServer2:

            def __init__(self, store=None):
                self.store = store or temp_folder(path_with_spaces=False)
                mkdir(self.store)
                self.fake_url = "http://fake%s.com" % str(uuid.uuid4()).replace("-", "")
                self.root_app = bottle.Bottle()
                self.app = TestApp(self.root_app)
                self._attach_to(self.root_app, self.store)

            @staticmethod
            def _attach_to(app, store):
                @app.route("/internet/<file>", method=["GET"])
                def get_internet_file(file):
                    return HTTPError(401, "You Are Not Allowed Here")

                @app.route("/downloader1/<file>", method=["GET"])
                def get_file(file):
                    return static_file(file, http_server_base_folder_backup1)

                @app.route("/downloader1/<file>", method=["PUT"])
                def put_file(file):
                    if file in os.listdir(http_server_base_folder_backup1):
                        return HTTPError(401, "You Are Not Allowed Here")
                    dest = os.path.join(http_server_base_folder_backup1, file)
                    with open(dest, 'wb') as f:
                        f.write(request.body.read())

                @app.route("/downloader2/<file>", method=["GET"])
                def get_file_2(file):
                    return static_file(file, http_server_base_folder_backup2)

        http_server = TestFileServer2()
        client.servers["file_server"] = http_server

        conanfile = textwrap.dedent(f"""
           from conan import ConanFile
           from conan.tools.files import download
           class Pkg2(ConanFile):
               name = "pkg"
               version = "1.0"
               def source(self):
                   download(self, "{http_server.fake_url}/internet/myfile.txt", "myfile.txt",
                            sha256="{sha256}")
           """)
        client.save_home({"global.conf": f"core.sources:download_cache={download_cache_folder}\n"
                          f"core.sources:download_urls=['{http_server.fake_url}/downloader1/', "
                          f"'origin', '{http_server.fake_url}/downloader2/']\n"
                          f"core.sources:upload_url={http_server.fake_url}/downloader1/\n"
                          "core.upload:retry=0\ncore.download:retry=0"})

        client.save({"conanfile.py": conanfile})
        client.run("create .")
        assert f"Sources for {http_server.fake_url}/internet/myfile.txt found in remote backup {http_server.fake_url}/downloader2/" in client.out
        # TODO: Check better message with Authentication error message
        assert "failed in 'origin'" in client.out

        # Now try to upload once to the first backup server. It's configured so it has write permissions but no overwrite
        client.run("upload * -c -r=default")
        upload_server_contents = os.listdir(http_server_base_folder_backup1)
        assert sha256 in upload_server_contents
        assert sha256 + ".json" in upload_server_contents

        client.run("remove * -c -r=default")

        client.run("upload * -c -r=default")
        assert f"Could not update summary '{sha256}.json' in backup sources server" in client.out

    def test_ok_when_origin_bad_sha256(self):
        http_server_base_folder_internet = os.path.join(self.file_server.store, "internet")
        http_server_base_folder_backup2 = os.path.join(self.file_server.store, "backup2")

        sha256 = "315f5bdb76d078c43b8ac0064e4a0164612b1fce77c869345bfc94c75894edd3"
        # This file's sha is not the one in the line above
        save(os.path.join(http_server_base_folder_internet, "myfile.txt"), "Bye, world!")
        save(os.path.join(http_server_base_folder_backup2, sha256), "Hello, world!")
        save(os.path.join(http_server_base_folder_backup2, sha256 + ".json"), '{"references": {}}')

        conanfile = textwrap.dedent(f"""
           from conan import ConanFile
           from conan.tools.files import download
           class Pkg2(ConanFile):
               name = "pkg"
               version = "1.0"
               def source(self):
                   download(self, "{self.file_server.fake_url}/internet/myfile.txt", "myfile.txt",
                            sha256="{sha256}")
           """)

        self.client.save_home(
            {"global.conf": f"core.sources:download_cache={self.download_cache_folder}\n"
                            f"core.sources:download_urls=['{self.file_server.fake_url}/backup1/', "
                            f"'origin', '{self.file_server.fake_url}/backup2/']\n"})

        self.client.save({"conanfile.py": conanfile})
        self.client.run("create .")
        assert f"Sources for {self.file_server.fake_url}/internet/myfile.txt found in remote backup {self.file_server.fake_url}/backup2/" in self.client.out
        assert "sha256 signature failed for" in self.client.out

    def test_export_then_upload_workflow(self):
        mkdir(os.path.join(self.download_cache_folder, "s"))

        http_server_base_folder_internet = os.path.join(self.file_server.store, "internet")
        http_server_base_folder_backup = os.path.join(self.file_server.store, "backup")

        sha256 = "315f5bdb76d078c43b8ac0064e4a0164612b1fce77c869345bfc94c75894edd3"
        save(os.path.join(http_server_base_folder_internet, "myfile.txt"), "Hello, world!")

        conanfile = textwrap.dedent(f"""
           from conan import ConanFile
           from conan.tools.files import download
           class Pkg2(ConanFile):
               name = "pkg"
               version = "1.0"
               def source(self):
                   download(self, "{self.file_server.fake_url}/internet/myfile.txt", "myfile.txt",
                            sha256="{sha256}")
           """)

        self.client.save_home(
            {"global.conf": f"core.sources:download_cache={self.download_cache_folder}\n"
                            f"core.sources:download_urls=['{self.file_server.fake_url}/backup/', 'origin']\n"
                            f"core.sources:upload_url={self.file_server.fake_url}/backup/"})

        self.client.save({"conanfile.py": conanfile})
        self.client.run("export .")
        self.client.run("upload * -c -r=default")
        # This used to crash because we were trying to list a missing dir if only exports were made
        assert "[Errno 2] No such file or directory" not in self.client.out
        self.client.run("create .")
        self.client.run("upload * -c -r=default")
        assert sha256 in os.listdir(http_server_base_folder_backup)
        self.client.run("cache clean * -bs")
        backups = os.listdir(os.path.join(self.download_cache_folder, "s"))
        assert len(backups) == 0

    def test_export_then_upload_recipe_only_workflow(self):
        http_server_base_folder_internet = os.path.join(self.file_server.store, "internet")

        sha256 = "315f5bdb76d078c43b8ac0064e4a0164612b1fce77c869345bfc94c75894edd3"
        save(os.path.join(http_server_base_folder_internet, "myfile.txt"), "Hello, world!")

        conanfile = textwrap.dedent(f"""
           from conan import ConanFile
           from conan.tools.files import download
           class Pkg2(ConanFile):
               name = "pkg"
               version = "1.0"
               def source(self):
                   download(self, "{self.file_server.fake_url}/internet/myfile.txt", "myfile.txt",
                            sha256="{sha256}")
           """)

        self.client.save_home(
            {"global.conf": f"core.sources:download_cache={self.download_cache_folder}\n"
                            f"core.sources:download_urls=['{self.file_server.fake_url}/backup/', 'origin']\n"
                            f"core.sources:upload_url={self.file_server.fake_url}/backup/"})

        self.client.save({"conanfile.py": conanfile})
        self.client.run("export .")
        exported_rev = self.client.exported_recipe_revision()
        self.client.run("upload * --only-recipe -c -r=default")
        # This second run used to crash because we thought there would be some packages always
        self.client.run("upload * --only-recipe -c -r=default")
        # Ensure we are testing for an already uploaded recipe
        assert f"Recipe 'pkg/1.0#{exported_rev}' already in server, skipping upload" in self.client.out

    @pytest.mark.parametrize("unknown", [True, False])
    def test_source_then_upload_workflow(self, unknown):
        mkdir(os.path.join(self.download_cache_folder, "s"))

        http_server_base_folder_internet = os.path.join(self.file_server.store, "internet")
        http_server_base_folder_backup = os.path.join(self.file_server.store, "backup")

        sha256 = "315f5bdb76d078c43b8ac0064e4a0164612b1fce77c869345bfc94c75894edd3"
        save(os.path.join(http_server_base_folder_internet, "myfile.txt"), "Hello, world!")

        conanfile = textwrap.dedent(f"""
           from conan import ConanFile
           from conan.tools.files import download
           class Pkg2(ConanFile):
               def source(self):
                   download(self, "{self.file_server.fake_url}/internet/myfile.txt", "myfile.txt",
                            sha256="{sha256}")
           """)

        self.client.save_home(
            {"global.conf": f"core.sources:download_cache={self.download_cache_folder}\n"
                            f"core.sources:download_urls=['{self.file_server.fake_url}/backup/', 'origin']\n"
                            f"core.sources:upload_url={self.file_server.fake_url}/backup/"})

        self.client.save({"conanfile.py": conanfile})
        ref_args = "" if unknown else "--name foo --version 1.0"
        reference_key = "unknown" if unknown else "foo/1.0"
        self.client.run(f"source . {ref_args}")
        content = json.loads(load(os.path.join(self.download_cache_folder, "s", sha256 + ".json")))
        assert len(content["references"][reference_key]) == 1

        self.client.run("cache backup-upload")
        # This used to crash because we were trying to list a missing dir if only exports were made
        assert "[Errno 2] No such file or directory" not in self.client.out
        assert sha256 in os.listdir(http_server_base_folder_backup)

    def test_backup_source_corrupted_download_handling(self):
        http_server_base_folder_internet = os.path.join(self.file_server.store, "internet")

        sha256 = "315f5bdb76d078c43b8ac0064e4a0164612b1fce77c869345bfc94c75894edd3"
        save(os.path.join(http_server_base_folder_internet, "myfile.txt"), "Hello, world!")

        conanfile = textwrap.dedent(f"""
           from conan import ConanFile
           from conan.tools.files import download
           class Pkg2(ConanFile):
               def source(self):
                   download(self, "{self.file_server.fake_url}/internet/myfile.txt", "myfile.txt",
                            sha256="{sha256}")
           """)

        self.client.save_home(
            {"global.conf": f"core.sources:download_cache={self.download_cache_folder}"})

        self.client.save({"conanfile.py": conanfile})
        self.client.run("source .")

        # Now simulate a dirty download, the file is there but the sha256 is wrong
        save(os.path.join(self.download_cache_folder, "s", sha256), "Bye bye, world!")

        # Now try to source again, it should eat it up
        self.client.run("source .")

    @pytest.mark.parametrize("exception", [Exception, ConanException])
    @pytest.mark.parametrize("upload", [True, False])
    def test_backup_source_dirty_download_handle(self, exception, upload):
        def custom_download(this, *args, **kwargs):
            raise exception()

        http_server_base_folder_internet = os.path.join(self.file_server.store, "internet")

        sha256 = "315f5bdb76d078c43b8ac0064e4a0164612b1fce77c869345bfc94c75894edd3"
        save(os.path.join(http_server_base_folder_internet, "myfile.txt"), "Hello, world!")

        conanfile = textwrap.dedent(f"""
           from conan import ConanFile
           from conan.tools.files import download
           class Pkg2(ConanFile):
               def source(self):
                   download(self, "{self.file_server.fake_url}/internet/myfile.txt", "myfile.txt",
                            sha256="{sha256}")
           """)

        self.client.save_home(
            {"global.conf": f"core.sources:download_cache={self.download_cache_folder}\n"
                            f"tools.files.download:retry=0"})

        self.client.save({"conanfile.py": conanfile})

        with mock.patch("conan.internal.rest.file_downloader.FileDownloader._download_file",
                        custom_download):
            self.client.run("source .", assert_error=True)
            # The mock does not actually download a file, let's add it for the test
            save(os.path.join(self.download_cache_folder, "s", sha256), "__corrupted download__")
            # Check that the dirty file was not removed.
            # This check should go away once we refactor dirty handling
            assert os.path.exists(os.path.join(self.download_cache_folder, "s", f"{sha256}.dirty"))

        if upload:
            self.client.run("cache backup-upload")
        else:
            self.client.run("source .")
        assert f"{sha256} is dirty, removing it" in self.client.out


    @pytest.mark.parametrize("exception", [Exception, ConanException])
    def test_backup_source_upload_when_dirty(self, exception):
        def custom_download(this, *args, **kwargs):  # noqa
            raise exception()

        mkdir(os.path.join(self.download_cache_folder, "s"))
        http_server_base_folder_internet = os.path.join(self.file_server.store, "internet")

        sha256 = "315f5bdb76d078c43b8ac0064e4a0164612b1fce77c869345bfc94c75894edd3"
        save(os.path.join(http_server_base_folder_internet, "myfile.txt"), "Hello, world!")

        conanfile = textwrap.dedent(f"""
           from conan import ConanFile
           from conan.tools.files import download
           class Pkg2(ConanFile):
               def source(self):
                   download(self, "{self.file_server.fake_url}/internet/myfile.txt", "myfile.txt",
                            sha256="{sha256}")
           """)

        self.client.save_home(
            {"global.conf": f"core.sources:download_cache={self.download_cache_folder}\n"
                            "tools.files.download:retry=0\n"
                            f"core.sources:upload_url={self.file_server.fake_url}/backup/\n"})

        self.client.save({"conanfile.py": conanfile})
        with mock.patch("conan.internal.rest.file_downloader.FileDownloader._download_file",
                        custom_download):
            self.client.run("source .", assert_error=True)

        # A .dirty file was created, now try to source again,
        # it should detect the dirty download and clean it, but not re-source it
        self.client.run("export . --name=pkg2 --version=1.0")
        self.client.run("upload * -c -r=default")
        assert "No backup sources files to upload" in self.client.out
        assert sha256 + ".dirty" not in os.listdir(os.path.join(self.download_cache_folder, "s"))



import os
import textwrap

from conan.api.model import RecipeReference
from conan.internal.util.files import load
from conan.test.utils.tools import TestClient, NO_SETTINGS_PACKAGE_ID


def test_package_python_files():
    client = TestClient(default_server_user=True)

    conanfile = textwrap.dedent("""
        from conan import ConanFile
        from conan.tools.files import copy
        class Pkg(ConanFile):
            exports_sources = "*"
            def package(self):
                copy(self, "*", self.source_folder, self.package_folder)
        """)
    client.save({"conanfile.py": conanfile,
                 "myfile.pyc": "",
                 "myfile.pyo": "",
                 ".DS_Store": ""})
    client.run("create . --name=pkg --version=0.1")
    ref_layout = client.exported_layout()
    pkg_layout = client.created_layout()
    export = ref_layout.export()
    export_sources = ref_layout.export_sources()
    assert os.path.isfile(os.path.join(export_sources, "myfile.pyc"))
    assert os.path.isfile(os.path.join(export_sources, "myfile.pyo"))
    assert os.path.isfile(os.path.join(export_sources, ".DS_Store"))
    manifest = load(os.path.join(export, "conanmanifest.txt"))
    assert "myfile.pyc" in manifest
    assert "myfile.pyo" in manifest
    assert ".DS_Store" not in manifest
    pkg_folder = pkg_layout.package()
    assert os.path.isfile(os.path.join(pkg_folder, "myfile.pyc"))
    assert os.path.isfile(os.path.join(pkg_folder, "myfile.pyo"))
    assert os.path.isfile(os.path.join(pkg_folder, ".DS_Store"))
    manifest = load(os.path.join(pkg_folder, "conanmanifest.txt"))
    assert "myfile.pyc" in manifest
    assert "myfile.pyo" in manifest
    assert ".DS_Store" not in manifest

    client.run("upload * -r=default --confirm")
    client.run("remove * -c")
    client.run("download pkg/0.1#*:* -r default")

    # The download will be in a different pkg folder now.
    ref = RecipeReference.loads("pkg/0.1")
    pref = client.get_latest_package_reference(ref, NO_SETTINGS_PACKAGE_ID)
    pkg_folder = client.get_latest_pkg_layout(pref).package()

    assert os.path.isfile(os.path.join(export_sources, "myfile.pyc"))
    assert os.path.isfile(os.path.join(export_sources, "myfile.pyo"))
    assert not os.path.isfile(os.path.join(export_sources, ".DS_Store"))
    manifest = load(os.path.join(export, "conanmanifest.txt"))
    assert "myfile.pyc" in manifest
    assert "myfile.pyo" in manifest
    assert ".DS_Store" not in manifest
    assert os.path.isfile(os.path.join(pkg_folder, "myfile.pyc"))
    assert os.path.isfile(os.path.join(pkg_folder, "myfile.pyo"))
    assert not os.path.isfile(os.path.join(pkg_folder, ".DS_Store"))
    manifest = load(os.path.join(pkg_folder, "conanmanifest.txt"))
    assert "myfile.pyc" in manifest
    assert "myfile.pyo" in manifest
    assert ".DS_Store" not in manifest


def test_package_empty_folders():
    c = TestClient(default_server_user=True)

    conanfile = textwrap.dedent("""
        import os
        from conan import ConanFile
        class Pkg(ConanFile):
            def package(self):
                os.mkdir(os.path.join(self.package_folder, "empty_folder"))
        """)
    c.save({"conanfile.py": conanfile})
    c.run("create . --name=pkg --version=0.1")
    pkg_folder = c.created_layout().package()
    assert os.path.isdir(os.path.join(pkg_folder, "empty_folder"))
    manifest = load(os.path.join(pkg_folder, "conanmanifest.txt"))
    assert "empty_folder" not in manifest

    c.run("upload * -r=default --confirm")
    c.run("remove * -c")
    c.run("download pkg/0.1#*:* -r default")

    # The download will be in a different pkg folder now.
    ref = RecipeReference.loads("pkg/0.1")
    pref = c.get_latest_package_reference(ref, NO_SETTINGS_PACKAGE_ID)
    pkg_folder = c.get_latest_pkg_layout(pref).package()
    assert os.path.isdir(os.path.join(pkg_folder, "empty_folder"))
    manifest = load(os.path.join(pkg_folder, "conanmanifest.txt"))
    assert "empty_folder" not in manifest



import textwrap

import pytest

from conan.test.assets.genconanfile import GenConanfile
from conan.test.utils.tools import TestClient, NO_SETTINGS_PACKAGE_ID


class TestInvalidConfiguration:
    """
    ConanInvalidConfiguration without a binary fall backs, result in errors
    """
    conanfile = textwrap.dedent("""
        from conan import ConanFile
        from conan.errors import ConanInvalidConfiguration

        class Conan(ConanFile):
            settings = "os"

            def validate(self):
                if self.info.settings.os == "Windows":
                    raise ConanInvalidConfiguration("Package does not work in Windows!")
       """)
    linux_package_id = "9a4eb3c8701508aa9458b1a73d0633783ecc2270"
    invalid = "Invalid"

    @pytest.fixture(scope="class")
    def client(self):
        client = TestClient()
        client.save({"pkg/conanfile.py": self.conanfile})
        client.run("create pkg --name=pkg --version=0.1 -s os=Linux")
        return client

    def test_invalid(self, client):
        conanfile_consumer = GenConanfile().with_requires("pkg/0.1").with_settings("os")
        client.save({"consumer/conanfile.py": conanfile_consumer})
        client.run("install consumer -s os=Windows", assert_error=True)
        assert "pkg/0.1: Invalid: Package does not work in Windows!" in client.out

    def test_invalid_info(self, client):
        """
        the conan info command does not raise, but it outputs info
        """
        conanfile_consumer = GenConanfile().with_requires("pkg/0.1").with_settings("os")
        client.save({"consumer/conanfile.py": conanfile_consumer})
        client.run("graph info consumer -s os=Windows")
        assert "binary: Invalid" in client.out

    def test_valid(self, client):
        conanfile_consumer = GenConanfile().with_requires("pkg/0.1").with_settings("os")
        client.save({"consumer/conanfile.py": conanfile_consumer})
        client.run("install consumer -s os=Linux")
        client.assert_listed_binary({"pkg/0.1": (self.linux_package_id, "Cache")})
        assert "pkg/0.1: Already installed!" in client.out

    def test_invalid_build_require(self, client):
        conanfile_consumer = GenConanfile().with_tool_requires("pkg/0.1").with_settings("os")
        client.save({"consumer/conanfile.py": conanfile_consumer})
        client.run("install consumer -s:h os=Windows -s:b os=Windows", assert_error=True)
        assert "pkg/0.1: Invalid: Package does not work in Windows!" in client.out

    def test_valid_build_require_two_profiles(self, client):
        conanfile_consumer = GenConanfile().with_tool_requires("pkg/0.1").with_settings("os")
        client.save({"consumer/conanfile.py": conanfile_consumer})
        client.run("install consumer -s:b os=Linux -s:h os=Windows")
        client.assert_listed_binary({"pkg/0.1": (self.linux_package_id, "Cache")}, build=True)
        assert "pkg/0.1: Already installed!" in client.out


class TestErrorConfiguration(TestInvalidConfiguration):
    """
    A configuration error is unsolvable, even if a binary exists
    """
    conanfile = textwrap.dedent("""
        from conan import ConanFile
        from conan.errors import ConanInvalidConfiguration

        class Conan(ConanFile):
            settings = "os"

            def validate(self):
                if self.info.settings.os == "Windows":
                    raise ConanInvalidConfiguration("Package does not work in Windows!")

            def package_id(self):
                del self.info.settings.os
        """)
    linux_package_id = NO_SETTINGS_PACKAGE_ID
    invalid = "ConfigurationError"


class TestErrorConfigurationCompatible(TestInvalidConfiguration):
    """
    A configuration error is unsolvable, even if a binary exists
    """
    conanfile = textwrap.dedent("""
        from conan import ConanFile
        from conan.errors import ConanInvalidConfiguration

        class Conan(ConanFile):
            settings = "os"

            def validate(self):
                if self.info.settings.os == "Windows":
                    raise ConanInvalidConfiguration("Package does not work in Windows!")

            def compatibility(self):
               if self.settings.os == "Windows":
                   return [{"settings": [("os", "Linux")]}]
        """)
    linux_package_id = "9a4eb3c8701508aa9458b1a73d0633783ecc2270"
    invalid = "ConfigurationError"


class TestInvalidBuildPackageID:
    """
    ConanInvalidBuildConfiguration will not block if setting is removed from package_id
    """
    conanfile = textwrap.dedent("""
       from conan import ConanFile
       from conan.errors import ConanInvalidConfiguration

       class Conan(ConanFile):
           settings = "os"

           def validate_build(self):
               if self.settings.os == "Windows":
                   raise ConanInvalidConfiguration("Package does not work in Windows!")

           def package_id(self):
               del self.info.settings.os
        """)
    linux_package_id = NO_SETTINGS_PACKAGE_ID
    windows_package_id = NO_SETTINGS_PACKAGE_ID

    @pytest.fixture(scope="class")
    def client(self):
        client = TestClient()
        client.save({"pkg/conanfile.py": self.conanfile})
        client.run("create pkg --name=pkg --version=0.1 -s os=Linux")
        return client

    def test_valid(self, client):
        conanfile_consumer = GenConanfile().with_requires("pkg/0.1").with_settings("os")
        client.save({"consumer/conanfile.py": conanfile_consumer})
        client.run("install consumer -s os=Windows")
        client.assert_listed_binary({"pkg/0.1": (self.linux_package_id, "Cache")})
        assert "pkg/0.1: Already installed!" in client.out

        client.run("install consumer -s os=Linux")
        client.assert_listed_binary({"pkg/0.1": (self.linux_package_id, "Cache")})
        assert "pkg/0.1: Already installed!" in client.out

    def test_invalid_try_build(self, client):
        conanfile_consumer = GenConanfile().with_requires("pkg/0.1").with_settings("os")
        client.save({"consumer/conanfile.py": conanfile_consumer})
        client.run("install consumer -s os=Windows --build='*'", assert_error=True)
        # Only when trying to build, it will try to build the Windows one
        client.assert_listed_binary({"pkg/0.1": (self.windows_package_id, "Invalid")})
        assert "Package does not work in Windows!" in client.out

    def test_valid_build_require_two_profiles(self, client):
        conanfile_consumer = GenConanfile().with_tool_requires("pkg/0.1").with_settings("os")
        client.save({"consumer/conanfile.py": conanfile_consumer})
        client.run("install consumer -s:b os=Linux -s:h os=Windows")
        client.assert_listed_binary({"pkg/0.1": (self.linux_package_id, "Cache")}, build=True)
        assert "pkg/0.1: Already installed!" in client.out

        client.run("install consumer -s:b os=Windows -s:h os=Windows")
        client.assert_listed_binary({"pkg/0.1": (self.linux_package_id, "Cache")}, build=True)
        assert "pkg/0.1: Already installed!" in client.out


class TestInvalidBuildCompatible(TestInvalidBuildPackageID):
    """
    ConanInvalidBuildConfiguration will not block if compatible_packages fallback
    """
    conanfile = textwrap.dedent("""
       from conan import ConanFile
       from conan.errors import ConanInvalidConfiguration

       class Conan(ConanFile):
           settings = "os"

           def validate_build(self):
               if self.settings.os == "Windows":
                   raise ConanInvalidConfiguration("Package does not work in Windows!")

           def compatibility(self):
               if self.settings.os == "Windows":
                   return [{"settings": [("os", "Linux")]}]
       """)
    linux_package_id = "9a4eb3c8701508aa9458b1a73d0633783ecc2270"
    windows_package_id = "ebec3dc6d7f6b907b3ada0c3d3cdc83613a2b715"



import json
import os
from pathlib import Path
import textwrap

import pytest

from conan.test.assets.genconanfile import GenConanfile

from conan.test.utils.tools import TestClient
from conan.internal.util.files import load, save

conanfile_dep = textwrap.dedent("""
    import os
    from conan import ConanFile
    from conan.tools.files import save, copy

    class TestConan(ConanFile):
        name = "dep"
        version = "1.0"
        def package(self):
            save(self, os.path.join(self.package_folder, "file.txt"), "Hello World!")
            save(self, os.path.join(self.package_folder, "file2.txt"), "Hello World 2!")

        def finalize(self):
            self.output.info(f"Running finalize method in {self.package_folder}")
            copy(self, "file.txt", src=self.immutable_package_folder, dst=self.package_folder)
            save(self, os.path.join(self.package_folder, "finalized.txt"), "finalized file")

        def package_info(self):
            self.output.info(f"Running package_info method in {self.package_folder}")
    """)


class TestBasicLocalFlows:

    @pytest.fixture
    def client(self):
        tc = TestClient(light=True)
        tc.save({"dep/conanfile.py": conanfile_dep})
        tc.run("export dep")
        return tc

    def test_basic_finalize_method(self, client):
        client.run("create dep")
        layout = client.created_layout()
        assert layout.package().endswith("p")
        assert f"Package folder {layout.package()}" in client.out
        assert f"Running finalize method in {layout.finalize()}" in client.out
        assert f"Running package_info method in {layout.finalize()}" in client.out
        client.run("install --requires=dep/1.0")
        assert f"Running package_info method in {layout.finalize()}" in client.out
        # Only issue is that the PackageLayout has no idea about the redirected package folder
        # So we have to know to check for it in tests, but oh well
        assert "finalized.txt" in os.listdir(layout.finalize())
        assert "finalized.txt" not in os.listdir(layout.package())

    def test_dependency_finalize_method(self, client):
        client.save({"app/conanfile.py": textwrap.dedent("""
                 from conan import ConanFile
                 class TestConan(ConanFile):
                     name = "app"
                     version = "1.0"
                     requires = "dep/1.0"
                     def generate(self):
                         self.output.info("Running generate method")
                         dep_pkg_folder = self.dependencies["dep"].package_folder
                         self.output.info(f"Dep package folder: {dep_pkg_folder}")
                 """)})
        client.run("create dep")
        dep_layout = client.created_layout()
        client.run("create app")
        assert f"Dep package folder: {dep_layout.package()}" not in client.out
        assert f"Dep package folder: {dep_layout.finalize()}" in client.out

    def test_no_non_info_access(self):
        client = TestClient(light=True)
        client.save({"conanfile.py": GenConanfile("dep", "1.0")
                     .with_finalize("self.output.info('settings.os: ' + self.settings.os)")})
        client.run("create .", assert_error=True)
        assert "'self.settings' access in 'finalize()' method is forbidden" in client.out

    def test_finalize_moves_from_package(self):
        client = TestClient(light=True)
        client.save({"conanfile.py": GenConanfile("dep", "1.0")
                     .with_import("from conan.tools.files import save, rename",
                                  "import os")
                     .with_option("move", [True, False])
                     .with_package('save(self, os.path.join(self.package_folder, "file.txt"), "Hello World!")',
                                   "save(self, os.path.join(self.package_folder, 'file2.txt'), 'Hello World 2!')")
                     # This is NOT allowed, moving from package to finalize is forbidden, only as test to ensure consistency
                     .with_finalize("rename(self, os.path.join(self.immutable_package_folder, 'file.txt'), os.path.join(self.package_folder, 'file.txt')) if self.info.options.move else None")})
        client.run("create . -o=dep/*:move=True")
        dep_moved_layout = client.created_layout()
        assert "file.txt" in os.listdir(dep_moved_layout.finalize())
        assert "file.txt" not in os.listdir(dep_moved_layout.package())

        client.run("create . -o=dep/*:move=False")
        dep_kept_layout = client.created_layout()
        assert "file.txt" not in os.listdir(dep_kept_layout.finalize())
        assert "file.txt" in os.listdir(dep_kept_layout.package())

        # Now we can check that the package_id is the same for both
        assert dep_moved_layout.reference.package_id != dep_kept_layout.reference.package_id

        # This now breaks if we try to cache check-integrity the moved package
        client.run(f"cache check-integrity {dep_moved_layout.reference}", assert_error=True)
        assert "There are corrupted artifacts" in client.out

        client.run(f"cache check-integrity {dep_kept_layout.reference}")
        assert "There are corrupted artifacts" not in client.out

    def test_cache_path_command(self, client):
        client.run("create dep")
        dep_layout = client.created_layout()
        pref = client.created_package_reference("dep/1.0")
        client.run(f"cache path {pref}")
        assert dep_layout.package() not in client.out
        assert dep_layout.finalize() in client.out

    def test_remove_deletes_correct_folders(self, client):
        client.run("create dep")
        dep_layout = client.created_layout()
        client.run("remove * -c")
        assert not os.path.exists(dep_layout.package())
        assert not os.path.exists(dep_layout.finalize())

    def test_save_restore_cache(self, client):
        # Not created in the cache, just exported, nothing breaks because there is not even a package there
        client.run("cache save *:*")
        # Check pkglist.json has not been created inside conan cache folder
        assert not any(Path(client.cache_folder).rglob("pgklist.json"))
        client.run("remove * -c")
        client.run("cache restore conan_cache_save.tgz")
        # Check the extracted pkglist does not persist in the cache after restore
        assert not any(Path(client.cache_folder).rglob("pgklist.json"))

        # Now create the package and then save/restore
        client.run("create dep")
        dep_layout = client.created_layout()
        client.run("cache save *:* -f=json", redirect_stdout="saved.json")
        saved = json.loads(client.load("saved.json"))
        pref = dep_layout.reference
        saved_pkg_folder = saved["Local Cache"]["dep/1.0"]["revisions"][pref.ref.revision]["packages"][pref.package_id]["revisions"][pref.revision]["package_folder"]
        assert saved_pkg_folder in dep_layout.package().replace("\\", "/")
        client.run("remove * -c")
        assert not os.path.exists(dep_layout.package())
        assert not os.path.exists(dep_layout.finalize())
        client.run("cache restore conan_cache_save.tgz")
        client.run(f"cache path {dep_layout.reference}")
        package_folder = client.out.strip()

        # The finalize() folder does not exist as restoring is not considered usage, so it never runs
        # so this is just the immutable package_folder
        assert "finalized.txt" not in os.listdir(package_folder)

        # But as soon as you call conan finalize, finalize() is called and so it's used,
        # so package_folder will be the finalize folder
        client.run("install --requires=dep/1.0")
        client.run(f"cache path {dep_layout.reference}")
        package_folder = client.out.strip()
        assert "finalized.txt" in os.listdir(package_folder)

    def test_graph_info_output(self, client):
        client.run("create dep")
        dep_layout = client.created_layout()
        client.run("install --requires=dep/1.0 -f=json", redirect_stdout="finalize.json")
        finalize_output = json.loads(client.load("finalize.json"))
        assert finalize_output["graph"]["nodes"]["1"]["package_folder"] == dep_layout.finalize()
        assert finalize_output["graph"]["nodes"]["1"]["immutable_package_folder"] == dep_layout.package()

    def test_create_pkglist_output(self, client):
        client.run("create dep -f=json", redirect_stdout="created.json")
        created_pkgid = client.created_package_id("dep/1.0")
        client.run("list --graph=created.json --graph-binaries=build")
        assert created_pkgid in client.out

    def test_vendorized_basic(self, client):
        client.run("create dep")
        client.save({"vendor/conanfile.py": GenConanfile("vendor", "1.0")
                     .with_import("from conan.tools.files import copy")
                     .with_class_attribute("vendor=True")
                     .with_requires("dep/1.0")
                     .with_package("copy(self, 'file.txt', src=self.dependencies['dep'].package_folder, dst=self.package_folder)",
                                   "copy(self, 'finalized.txt', src=self.dependencies['dep'].package_folder, dst=self.package_folder)",
                                   "copy(self, 'file2.txt', src=self.dependencies['dep'].immutable_package_folder, dst=self.package_folder)")})
        client.run("create vendor")
        vendor_layout = client.created_layout()
        assert "file.txt" in os.listdir(vendor_layout.package())
        assert "finalized.txt" in os.listdir(vendor_layout.package())
        assert "file2.txt" in os.listdir(vendor_layout.package())

    def test_check_integrity(self, client):
        client.run("create dep")
        dep_layout = client.created_layout()
        client.run(f"cache check-integrity {dep_layout.reference}")
        assert "There are corrupted artifacts" not in client.out
        # Even if we re-change the finalize folder contents, it should still be fine
        save(os.path.join(dep_layout.finalize(), "finalized.txt"), "Modified!")
        client.run(f"cache check-integrity {dep_layout.reference}")
        assert "There are corrupted artifacts" not in client.out
        # But as soon as we change the package, it should still fail like a normal package would
        save(os.path.join(dep_layout.package(), "file.txt"), "Modified!")
        client.run(f"cache check-integrity {dep_layout.reference}", assert_error=True)
        assert "There are corrupted artifacts" in client.out

    @pytest.mark.parametrize("with_finalize_method", [True, False])
    def test_access_immutable_from_consumer(self, client, with_finalize_method):
        if not with_finalize_method:
            client.save({"dep/conanfile.py": GenConanfile("dep", "1.0")})
        client.save({"app/conanfile.py": GenConanfile("app", "1.0")
                     .with_requires("dep/1.0")
                     .with_package("dep = self.dependencies['dep/1.0']",
                                   "self.output.info(f'Immutable package: {dep.immutable_package_folder}')",
                                   # TODO: Think about if we want this interface
                                   # "self.output.info(f'finalize: {dep.finalize_folder}')",
                                   "self.output.info(f'Package: {dep.package_folder}')")})
        client.run("create dep")
        dep_layout = client.created_layout()
        client.run("create app")
        assert f"app/1.0: Immutable package: {dep_layout.package()}" in client.out
        # assert f"app/1.0: finalize: {dep_layout.finalize()}" in client.out
        if with_finalize_method:
            assert f"app/1.0: Package: {dep_layout.finalize()}" in client.out
        else:
            assert f"app/1.0: Package: {dep_layout.package()}" in client.out

    def test_cache_modification_of_custom_conf_based_on_settings(self):
        tc = TestClient(light=True)
        tc.save({"conanfile.py": GenConanfile("dep", "1.0")
                .with_import("from conan.tools.files import save",
                             "import os")
                .with_option("myoption", [True, False])
                .with_option("otheroption", [True, False])
                .with_default_option("myoption", False)
                .with_default_option("otheroption", False)
                .with_setting("os")
                .with_package_id("del self.info.options.myoption")
                .with_finalize("save(self, os.path.join(self.package_folder, 'file.txt'), 'Hello World!')",
                              "save(self, os.path.join(self.package_folder, 'os.conf'), str(self.info.settings.os))",
                              "save(self, os.path.join(self.package_folder, 'option.conf'), str(self.info.options.get_safe('myoption')))",
                              "save(self, os.path.join(self.package_folder, 'otheroption.conf'), str(self.info.options.otheroption))")})
        tc.run("create . -s=os=Linux -o=&:myoption=True -o=&:otheroption=True")
        layout = tc.created_layout()
        assert "file.txt" in os.listdir(layout.finalize())
        assert tc.load(os.path.join(layout.finalize(), "os.conf")) == "Linux"
        # This is problematic, it means that the mapping for finalize() and package_id would not be 1:1 and could be outdated
        assert tc.load(os.path.join(layout.finalize(), "option.conf")) == "None"
        assert tc.load(os.path.join(layout.finalize(), "otheroption.conf")) == "True"


class TestToolRequiresFlows:
    def test_tool_requires(self):
        tc = TestClient(light=True)
        tc.save({"dep/conanfile.py": textwrap.dedent("""
            import os
            from conan import ConanFile
            from conan.tools.files import save, copy

            class TestConan(ConanFile):
                name = "dep"
                version = "1.0"
                package_type = "application"
                def package(self):
                    save(self, os.path.join(self.package_folder, "bin", "executable.txt"), "Base")

                def finalize(self):
                    self.output.info(f"Running finalize method in {self.package_folder}")
                    copy(self, "*", src=self.immutable_package_folder, dst=self.package_folder)
                    save(self, os.path.join(self.package_folder, "bin", "finalized.txt"), "finalized file")

                def package_info(self):
                    self.output.info(f"Running package_info method in {self.package_folder}")
                    self.cpp_info.bindirs = ["bin"]

            """), "app/conanfile.py": textwrap.dedent("""
            from conan import ConanFile
            import os

            class TestConan(ConanFile):
                name = "app"
                version = "1.0"

                def build_requirements(self):
                    self.tool_requires("dep/1.0")

                def build(self):
                    self.output.info("Running build method")
                    bindir = self.dependencies.build['dep'].cpp_info.bindir
                    self.output.info(f"Dep bindir: {bindir}")
                    self.output.info(f"Is finalized? {os.path.exists(os.path.join(bindir, 'finalized.txt'))}")
            """)})
        tc.run("create dep --build-require")
        dep_layout = tc.created_layout()
        tc.run("create app")
        # This fails. cpp_info is using the original package folder to construct the final path
        assert f"Dep bindir: {dep_layout.finalize()}" in tc.out
        assert "app/1.0: Is finalized? True" in tc.out

    def test_test_package_uses_created_tool_which_modifies_pkgfolder(self):
        tc = TestClient(light=True)
        tc.save({"conanfile.py": GenConanfile("app", "1.0")
                .with_import("from conan.tools.files import save")
                .with_package_type("application")
                .with_package("save(self, 'file.txt', 'Hello World!')")
                .with_package_info({"bindirs": ["bin"]}, {})
                .with_finalize("save(self, 'finalized.txt', 'finalized file')"),
                 "test_package/conanfile.py": GenConanfile()
                .with_import("from conan.tools.files import save",
                             "import os")
                .with_test_reference_as_build_require()
                .with_test("bindir = self.dependencies.build[self.tested_reference_str].cpp_info.bindir",
                           "self.output.info(f'Bindir: {bindir}')",
                           "save(self, os.path.join(bindir, '__pycache__.pyc'), 'Test file')")})
        tc.run("create . --build-require")
        app_layout = tc.created_layout()
        assert f"Bindir: {os.path.join(app_layout.finalize(), 'bin')}" in tc.out
        tc.run(f"cache check-integrity {app_layout.reference}")
        assert "There are corrupted artifacts" not in tc.out


class TestRemoteFlows:

    @pytest.fixture
    def client(self):
        tc = TestClient(light=True, default_server_user=True)
        tc.save({"dep/conanfile.py": conanfile_dep})
        tc.run("export dep")
        return tc

    def test_remote_upload_finalize_method(self, client):
        client.run("create dep")
        created_pref = client.created_package_reference("dep/1.0")
        client.run("upload * -r=default -c")

        # Only the package folder is uploaded, not the finalize folder
        uploaded_pref_path = client.servers["default"].test_server.server_store.package(created_pref)
        manifest_contents = load(os.path.join(uploaded_pref_path, "conanmanifest.txt"))
        assert "file.txt" in manifest_contents
        assert "finalized.txt" not in manifest_contents

        client.run("remove * -c")
        client.run(f"download {created_pref} -r=default")
        downloaded_pref_layout = client.get_latest_pkg_layout(created_pref)
        assert "file.txt" in os.listdir(downloaded_pref_layout.package())
        # Download is not an "usage" of the package, so no finalize() is yet executed
        assert "finalized.txt" not in os.listdir(downloaded_pref_layout.package())
        assert not os.path.exists(os.path.join(downloaded_pref_layout.finalize()))

        client.run(f"cache path {created_pref}")
        package_folder = client.out.strip()
        assert package_folder == downloaded_pref_layout.package()
        assert package_folder.endswith("p")
        # Now this finalize will run the finalize() method
        client.run("install --requires=dep/1.0")
        assert f"Running finalize method in {downloaded_pref_layout.finalize()}" in client.out

        client.run("remove * -c")
        client.run("install --requires=dep/1.0 -r=default")
        assert "dep/1.0: Calling finalize()"
        assert f"Running finalize method in {downloaded_pref_layout.finalize()}" in client.out

    def test_upload_verify_integrity(self, client):
        client.run("create dep")
        dep_layout = client.created_layout()
        client.run("upload * -r=default -c --check")
        assert f"dep/1.0#{dep_layout.reference.ref.revision}:{dep_layout.reference.package_id}" \
               f"#{dep_layout.reference.revision}: Integrity check: ok" in client.out
        assert "There are corrupted artifacts" not in client.out



import textwrap

from jinja2 import Template

from conan.test.utils.tools import TestClient
from conan.internal.util.files import save


class TestTestPackage:
    conanfile_tpl = Template(textwrap.dedent("""
        from conan import ConanFile
        import os
        from conan.tools.env import VirtualBuildEnv

        class Recipe(ConanFile):
            settings = "os"
            {{ build_requires|default("") }}

            {% if test %}
            def requirements(self):
                self.requires(self.tested_reference_str)
            {% endif %}

            {% raw %}
            def build(self):
                self.output.info(">> settings.os: {}".format(self.settings.os))
                self.output.info(">> settings_build.os: {}".format(self.settings_build.os))
                build_env = VirtualBuildEnv(self).vars()
                with build_env.apply():
                    self.output.info(">> tools.get_env('INFO'): {}".format(os.getenv("INFO")))

            def package_info(self):
                self.buildenv_info.define("INFO", "{}-{}".format(self.name, self.settings.os))

            def test(self):
                pass
            {% endraw %}

    """))

    conanfile_br = conanfile_tpl.render()
    conanfile = conanfile_tpl.render(build_requires='build_requires = "br1/version"')
    conanfile_test = conanfile_tpl.render(build_requires='build_requires = "br2/version"',
                                          test=True)

    settings_yml = textwrap.dedent("""
        os:
            Host:
            Build:
    """)

    def test_command(self):
        t = TestClient()
        save(t.paths.settings_path, self.settings_yml)
        t.save({'br.py': self.conanfile_br,
                'conanfile.py': self.conanfile,
                'test_package/conanfile.py': self.conanfile_test,
                'profile_host': '[settings]\nos=Host',
                'profile_build': '[settings]\nos=Build', })
        t.run("export br.py --name=br1 --version=version")
        # It is necessary to build first the test_package build_requires br2
        t.run("create br.py --name=br2 --version=version -tf=\"\" --build-require "
              "--profile:host=profile_host --profile:build=profile_build")

        t.run("create conanfile.py --name=name --version=version --build=missing"
              " --profile:host=profile_host --profile:build=profile_build")

        # Build requires are built in the 'build' context:
        assert "br1/version: >> settings.os: Build" in t.out
        assert "br1/version: >> settings_build.os: Build" in t.out

        # Package 'name' is built for the 'host' context (br1 as build_requirement)
        assert "name/version: >> settings.os: Host" in t.out
        assert "name/version: >> settings_build.os: Build" in t.out

        # Test_package is executed with the same profiles as the package itself
        assert "name/version (test package): >> settings.os: Host" in t.out
        assert "name/version (test package): >> settings_build.os: Build" in t.out

        t.run("test test_package/conanfile.py name/version@ "
              "--profile:host=profile_host --profile:build=profile_build")

        assert "name/version (test package): >> settings.os: Host" in t.out
        assert "name/version (test package): >> settings_build.os: Build" in t.out
        assert "name/version (test package): >> tools.get_env('INFO'): br2-Build" in t.out



import textwrap
import pytest

from conan.test.assets.genconanfile import GenConanfile
from conan.test.utils.tools import TestClient
new_value = "will_break_next"


@pytest.mark.tool("cmake", "3.27")
@pytest.mark.parametrize("generator", ["CMakeDeps", "CMakeConfigDeps"])
def test_package_info_extra_variables(generator):
    """ The dependencies can define extra variables to be used in CMake,
        but if the user is setting the cmake_extra_variables conf,
        those should have precedence.
    """
    client = TestClient()
    dep_conanfile = textwrap.dedent("""
        from conan import ConanFile

        class Pkg(ConanFile):
            name = "dep"
            version = "0.1"

            def package_info(self):
                self.cpp_info.set_property("cmake_extra_variables", {"FOO": 42})
    """)
    client.save({"dep/conanfile.py": dep_conanfile})
    client.run("create dep")

    cmakelists = textwrap.dedent("""
    cmake_minimum_required(VERSION 3.27)
    project(myproject CXX)
    find_package(dep CONFIG REQUIRED)
    message(STATUS "FOO=${FOO}")
    """)


    conanfile = textwrap.dedent(f"""
    from conan import ConanFile
    from conan.tools.cmake import CMake

    class Pkg(ConanFile):
        settings = "os", "arch", "compiler", "build_type"
        generators = "{generator}", "CMakeToolchain"
        requires = "dep/0.1"
        def build(self):
            cmake = CMake(self)
            cmake.configure()
    """)
    client.save({"CMakeLists.txt": cmakelists,
                 "conanfile.py": conanfile})
    conf = f"-c tools.cmake.cmakedeps:new={new_value}" if generator == "CMakeConfigDeps" else ""
    client.run(f"build . {conf} "
               """-c tools.cmake.cmaketoolchain:extra_variables="{'FOO': '9'}" """)

    assert "-- FOO=9" in client.out




from conan.api.subapi.new import NewAPI
from conan.internal.api.new.cmake_exe import cmake_exe_files
from conan.internal.api.new.cmake_lib import cmake_lib_files


conanws_yml = """\
packages:
  - path: liba
    ref: liba/0.1
  - path: libb
    ref: libb/0.1
  - path: app1
    ref: app1/0.1
"""

cmake = """\
cmake_minimum_required(VERSION 3.25)
project(monorepo CXX)

include(FetchContent)

function(add_project PACKAGE_NAME SUBFOLDER)
    message(STATUS "Adding project: ${PACKAGE_NAME}. Folder: ${SUBFOLDER}")
    FetchContent_Declare(
        ${PACKAGE_NAME}
        SOURCE_DIR ${CMAKE_CURRENT_LIST_DIR}/${SUBFOLDER}
        SYSTEM
        OVERRIDE_FIND_PACKAGE
    )
    FetchContent_MakeAvailable(${PACKAGE_NAME})
endfunction()

include(build/conanws_build_order.cmake)

foreach(pair ${CONAN_WS_BUILD_ORDER})
    string(FIND "${pair}" ":" pos)
    string(SUBSTRING "${pair}" 0 "${pos}" pkg)
    math(EXPR pos "${pos} + 1")  # Skip the separator
    string(SUBSTRING "${pair}" "${pos}" -1 folder)

    add_project(${pkg} ${folder})
    # This target should be defined in the liba/CMakeLists.txt, but we can fix it here
    get_target_property(target_type ${pkg} TYPE)
    if (NOT target_type STREQUAL "EXECUTABLE")
        add_library(${pkg}::${pkg} ALIAS ${pkg})
    endif()
endforeach()
"""

conanfile = r'''\
import json
from conan import Workspace
from conan import ConanFile
from conan.tools.files import save
from conan.tools.cmake import CMakeDeps, CMakeToolchain, cmake_layout


class MyWs(ConanFile):
    """ This is a special conanfile, used only for workspace definition of layout
    and generators. It shouldn't have requirements, tool_requirements. It shouldn't have
    build() or package() methods
    """
    settings = "os", "compiler", "build_type", "arch"

    def generate(self):
        deps = CMakeDeps(self)
        deps.generate()
        tc = CMakeToolchain(self)
        tc.generate()

    def layout(self):
        cmake_layout(self)


class Ws(Workspace):
    def root_conanfile(self):
        return MyWs

    def build_order(self, order):
        super().build_order(order)  # default behavior prints the build order
        pkglist = " ".join([f'{it["ref"].name}:{it["folder"]}' for level in order for it in level])
        save(self, "build/conanws_build_order.cmake", f"set(CONAN_WS_BUILD_ORDER {pkglist})")
'''

workspace_files = {"conanws.yml": conanws_yml,
                   "CMakeLists.txt": cmake,
                   "conanws.py": conanfile,
                   ".gitignore": "build"}
# liba
files = {f"liba/{k}": v for k, v in cmake_lib_files.items()}
workspace_files.update(files)
# libb
files = NewAPI.render(cmake_lib_files, {"requires": ["liba/0.1"], "name": "libb"})
files = {f"libb/{k}": v for k, v in files.items()}
workspace_files.update(files)
# app
files = NewAPI.render(cmake_exe_files, definitions={"name": "app1", "requires": ["libb/0.1"]})
files = {f"app1/{k}": v for k, v in files.items()}
workspace_files.update(files)



import os
import platform

import pytest

from conan.test.utils.tools import TestClient


@pytest.mark.skipif(platform.system() != "Windows", reason="Requires Windows")
def test_msbuild_lib_template():
    client = TestClient()
    client.run("new msbuild_lib -d name=hello -d version=0.1")
    # Local flow works
    client.run("install .")
    client.run("build .")

    assert os.path.isfile(os.path.join(client.current_folder, "x64", "Release", "hello.lib"))
    client.run("export-pkg .")
    package_folder = client.created_layout().package()
    assert os.path.exists(os.path.join(package_folder, "include", "hello.h"))
    assert os.path.exists(os.path.join(package_folder, "lib", "hello.lib"))

    # Create works
    client.run("create .")
    assert "hello/0.1: Hello World Release!" in client.out
    assert "hello/0.1: _MSC_VER191" in client.out

    client.run("create . -s build_type=Debug")
    assert "hello/0.1: Hello World Debug!" in client.out

    # FIXME: Create + shared DOESNT work fine, the proj is hardcoded to static
    # client.run("create . -o hello:shared=True")
    # assert "hello/0.1: Hello World Release!" in client.out


@pytest.mark.skipif(platform.system() != "Windows", reason="Requires Windows")
@pytest.mark.tool("visual_studio", "17")
def test_msbuild_lib_2022():
    # The template .vcxproj are MSBuildTools=15, so it won't work with older versions
    # 2022 Must have installed the v141 toolset too
    client = TestClient(path_with_spaces=False)
    client.run("new msbuild_lib -d name=hello -d version=0.1")

    # Create works
    client.run("create . -s compiler.version=191 -c tools.microsoft.msbuild:vs_version=17")
    assert "hello/0.1: Hello World Release!" in client.out
    # This is the default compiler.version=191 in conftest
    assert "Activating environment Visual Studio 17" in client.out
    assert "hello/0.1: _MSC_VER191" in client.out

    # Create works
    client.run("create . -s compiler.version=193")
    assert "hello/0.1: Hello World Release!" in client.out
    assert "Activating environment Visual Studio 17" in client.out
    assert "hello/0.1: _MSC_VER193" in client.out


@pytest.mark.skipif(platform.system() != "Windows", reason="Requires Windows")
def test_msbuild_exe_template():
    client = TestClient(path_with_spaces=False)
    client.run("new msbuild_exe -d name=greet -d version=0.1")
    # Local flow works
    client.run("install .")
    client.run("build .")

    # Create works
    client.run("create .")
    assert "greet/0.1: Hello World Release!" in client.out

    client.run("create . -s build_type=Debug")
    assert "greet/0.1: Hello World Debug!" in client.out



import json
import os
import platform
import stat
import textwrap

import pytest
from parameterized import parameterized

from conan.internal.model.manifest import FileTreeManifest
from conan.api.model import RecipeReference
from conan.internal.paths import CONANFILE, CONAN_MANIFEST
from conan.test.utils.tools import TestClient, GenConanfile
from conan.internal.util.files import load, save


class TestExportSettings:

    def test_export_without_full_reference(self):
        client = TestClient(light=True)
        client.save({"conanfile.py": GenConanfile()})
        client.run("export . --user=lasote --channel=stable", assert_error=True)
        assert "conanfile didn't specify name" in client.out

        client.save({"conanfile.py": GenConanfile("lib")})
        client.run("export . --user=lasote --channel=stable", assert_error=True)
        assert "conanfile didn't specify version" in client.out

        client.save({"conanfile.py": GenConanfile()})
        client.run("export . --name=lib --version=1.0 --user=lasote --channel=channel")
        assert "lib/1.0@lasote/channel: Exported" in client.out

        client.save({"conanfile.py": GenConanfile("lib", "1.0")})
        client.run("export . --user=lasote")
        assert "lib/1.0@lasote: Exporting package recipe" in client.out
        client.run("export . --channel=channel", assert_error=True)
        assert "Can't specify channel 'channel' without user" in client.out

        client.save({"conanfile.py": GenConanfile("lib", "1.0").with_class_attribute('channel = "channel"')})
        client.run("export .", assert_error=True)
        assert "Can't specify channel 'channel' without user" in client.out

        client.run("export . --user=user")
        assert "Exported: lib/1.0@user/channel" in client.out

        client.save({"conanfile.py": GenConanfile("lib", "1.0").with_class_attribute('user = "user"')})
        client.run("export .")
        assert "lib/1.0@user: Exported" in client.out
        client.run("export . --channel=channel")
        assert "lib/1.0@user/channel: Exported" in client.out


    def test_export_read_only(self):
        client = TestClient(light=True)
        conanfile = textwrap.dedent("""
            from conan import ConanFile
            class TestConan(ConanFile):
                name = "hello"
                version = "1.2"
                exports = "file1.txt"
                exports_sources = "file2.txt"
            """)
        client.save({CONANFILE: conanfile,
                     "file1.txt": "",
                     "file2.txt": ""})
        mode1 = os.stat(os.path.join(client.current_folder, "file1.txt")).st_mode
        mode2 = os.stat(os.path.join(client.current_folder, "file2.txt")).st_mode
        os.chmod(os.path.join(client.current_folder, "file1.txt"), mode1 & ~stat.S_IWRITE)
        os.chmod(os.path.join(client.current_folder, "file2.txt"), mode2 & ~stat.S_IWRITE)

        client.run("export . --user=lasote --channel=stable")

        layout = client.exported_layout()
        export_path = layout.export()
        export_src_path = layout.export_sources()

        assert load(os.path.join(export_path, "file1.txt")) == ""
        assert load(os.path.join(export_src_path, "file2.txt")) == ""
        with pytest.raises(IOError):
            save(os.path.join(export_path, "file1.txt"), "")
        with pytest.raises(IOError):
            save(os.path.join(export_src_path, "file2.txt"), "")

        os.chmod(os.path.join(client.current_folder, "file1.txt"), mode1 | stat.S_IWRITE)
        os.chmod(os.path.join(client.current_folder, "file2.txt"), mode2 | stat.S_IWRITE)
        client.save({CONANFILE: conanfile,
                     "file1.txt": "file1",
                     "file2.txt": "file2"})
        client.run("export . --user=lasote --channel=stable")

        layout = client.exported_layout()
        export_path = layout.export()
        export_src_path = layout.export_sources()

        assert load(os.path.join(export_path, "file1.txt")) == "file1"
        assert load(os.path.join(export_src_path, "file2.txt")) == "file2"
        client.run("install --requires=hello/1.2@lasote/stable --build=missing")
        assert "hello/1.2@lasote/stable: Generating the package" in client.out

        client.save({CONANFILE: conanfile,
                     "file1.txt": "",
                     "file2.txt": ""})
        os.chmod(os.path.join(client.current_folder, "file1.txt"), mode1 & ~stat.S_IWRITE)
        os.chmod(os.path.join(client.current_folder, "file2.txt"), mode2 & ~stat.S_IWRITE)
        client.run("export . --user=lasote --channel=stable")

        layout = client.exported_layout()
        export_path = layout.export()
        export_src_path = layout.export_sources()

        assert load(os.path.join(export_path, "file1.txt")) == ""
        assert load(os.path.join(export_src_path, "file2.txt")) == ""
        client.run("install --requires=hello/1.2@lasote/stable --build=hello*")
        assert "hello/1.2@lasote/stable: Generating the package" in client.out

    def test_code_parent(self):
        # when referencing the parent, the relative folder "sibling" will be kept
        base = """
import os
from conan import ConanFile
from conan.tools.files import copy
class TestConan(ConanFile):
    name = "hello"
    version = "1.2"
    def export(self):
        copy(self, "*.txt", src=os.path.join(self.recipe_folder, ".."),
             dst=self.export_folder)
"""
        for conanfile in (base, base.replace("../*.txt", "../sibling*")):
            client = TestClient(light=True)
            client.save({"recipe/conanfile.py": conanfile,
                         "sibling/file.txt": "Hello World!"})
            client.current_folder = os.path.join(client.current_folder, "recipe")
            client.run("export . --user=lasote --channel=stable")
            layout = client.exported_layout()
            export_path = layout.export()
            content = load(os.path.join(export_path, "sibling/file.txt"))
            assert "Hello World!" == content

    def test_code_sibling(self):
        # if provided a path with slash, it will use as a export base
        client = TestClient(light=True)
        conanfile = """
import os
from conan import ConanFile
from conan.tools.files import copy
class TestConan(ConanFile):
    name = "hello"
    version = "1.2"
    def export(self):
        copy(self, "*.txt", src=os.path.join(self.recipe_folder, "..", "sibling"),
             dst=self.export_folder)
"""
        files = {"recipe/conanfile.py": conanfile,
                 "sibling/file.txt": "Hello World!"}
        client.save(files)
        client.current_folder = os.path.join(client.current_folder, "recipe")
        client.run("export . --user=lasote --channel=stable")
        layout = client.exported_layout()
        export_path = layout.export()
        content = load(os.path.join(export_path, "file.txt"))
        assert "Hello World!" == content

    def test_code_several_sibling(self):
        # if provided a path with slash, it will use as a export base
        client = TestClient(light=True)
        conanfile = textwrap.dedent("""
            import os
            from conan import ConanFile
            from conan.tools.files import copy
            class TestConan(ConanFile):
                name = "hello"
                version = "1.2"
                # exports_sources = "../test/src/*", "../cpp/*", "../include/*"
                def export_sources(self):
                    copy(self, "*", src=os.path.join(self.recipe_folder, "..", "test", "src"),
                         dst=self.export_sources_folder)
                    copy(self, "*", src=os.path.join(self.recipe_folder, "..", "cpp"),
                         dst=self.export_sources_folder)
                    copy(self, "*", src=os.path.join(self.recipe_folder, "..", "include"),
                         dst=self.export_sources_folder)

            """)
        client.save({"recipe/conanfile.py": conanfile,
                     "test/src/file.txt": "Hello World!",
                     "cpp/file.cpp": "Hello World!",
                     "include/file.h": "Hello World!"})
        client.current_folder = os.path.join(client.current_folder, "recipe")
        client.run("export . --user=lasote --channel=stable")
        layout = client.exported_layout()
        export_path = layout.export_sources()
        assert sorted(['file.txt', 'file.cpp', 'file.h']) == sorted(os.listdir(export_path))

    @parameterized.expand([("myconanfile.py", ), ("Conanfile.py", )])
    def test_filename(self, filename):
        client = TestClient(light=True)
        client.save({filename: GenConanfile("hello", "1.2")})
        client.run("export %s --user=user --channel=stable" % filename)
        assert "hello/1.2@user/stable: Exported" in client.out
        layout = client.exported_layout()
        export_path = layout.export()
        conanfile = load(os.path.join(export_path, "conanfile.py"))
        assert "name = 'hello'" in conanfile
        manifest = load(os.path.join(export_path, "conanmanifest.txt"))
        assert 'conanfile.py: 5dc49e518e15f3889cb2e097ce4d1dff' in manifest

    def test_exclude_basic(self):
        client = TestClient(light=True)
        conanfile = """
from conan import ConanFile
class TestConan(ConanFile):
    name = "hello"
    version = "1.2"
    exports = "*.txt", "!*file1.txt"
    exports_sources = "*.cpp", "!*temp.cpp"
"""

        client.save({CONANFILE: conanfile,
                     "file.txt": "",
                     "file1.txt": "",
                     "file.cpp": "",
                     "file_temp.cpp": ""})
        client.run("export .")
        layout = client.exported_layout()
        export_path = layout.export()
        exports_sources_path = layout.export_sources()
        assert os.path.exists(os.path.join(export_path, "file.txt"))
        assert not os.path.exists(os.path.join(export_path, "file1.txt"))
        assert os.path.exists(os.path.join(exports_sources_path, "file.cpp"))
        assert not os.path.exists(os.path.join(exports_sources_path, "file_temp.cpp"))

    def test_exclude_folders(self):
        client = TestClient(light=True)
        client.save({CONANFILE: GenConanfile("hello", "1.2").with_exports("*.txt", "!*/temp/*"),
                     "file.txt": "",
                     "any/temp/file1.txt": "",
                     "other/sub/file2.txt": ""})
        client.run("export .")
        layout = client.exported_layout()
        export_path = layout.export()
        assert os.path.exists(os.path.join(export_path, "file.txt"))
        assert not os.path.exists(os.path.join(export_path, "any/temp/file1.txt"))
        assert os.path.exists(os.path.join(export_path, "other/sub/file2.txt"))


class TestExport:

    @pytest.fixture(autouse=True)
    def setup(self):
        self.client = TestClient(light=True)
        self.files = {"conanfile.py": GenConanfile("hello0", "0.1").with_exports("*"),
                      "main.cpp": "MyMain",
                      "CMakeLists.txt": "MyCmake",
                      "executable": "myexe"}
        self.ref = RecipeReference("hello0", "0.1", "lasote", "stable")
        self.client.save(self.files)
        self.client.run("export . --user=lasote --channel=stable")

    def test_basic(self):
        reg_path = self.client.exported_layout().export()
        manif = FileTreeManifest.load(reg_path)

        assert '%s: Exported' % str(self.ref) in self.client.out
        assert '%s: Exported to cache folder: %s' % (str(self.ref), reg_path) in self.client.out
        assert os.path.exists(reg_path)

        for name in list(self.files.keys()):
            assert os.path.exists(os.path.join(reg_path, name))

        expected_sums = {'CMakeLists.txt': '3cf710785270c7e98a30d4a90ea66492',
                         'conanfile.py': '5dbbe4328efa3342baba2a7ca961ede1',
                         'executable': 'db299d5f0d82f113fad627a21f175e59',
                         'main.cpp': 'd9c03c934a4b3b1670775c17c26f39e9'}
        assert expected_sums == manif.file_sums

    def test_case_sensitive(self):
        self.ref = RecipeReference("hello0", "0.1", "lasote", "stable")
        self.client.save({"conanfile.py": GenConanfile("hello0", "0.1").with_exports("*")})
        self.client.run("export . --user=lasote --channel=stable")
        assert "hello0/0.1@lasote/stable: Exported" in self.client.out

    def test_export_filter(self):
        self.client.save({CONANFILE: GenConanfile("openssl", "2.0.1")})
        self.client.run("export . --user=lasote --channel=stable")
        layout = self.client.exported_layout()
        export_path = layout.export()
        assert sorted(os.listdir(export_path)) == [CONANFILE, CONAN_MANIFEST]

        content = """
from conan import ConanFile

class OpenSSLConan(ConanFile):
    name = "openssl"
    version = "2.0.1"
    exports = ('*.txt', '*.h')
"""
        self.client.save({CONANFILE: content})
        self.client.run("export . --user=lasote --channel=stable")
        layout = self.client.exported_layout()
        export_path = layout.export()
        assert sorted(os.listdir(export_path)) == ['CMakeLists.txt', CONANFILE, CONAN_MANIFEST]

        # Now exports being a list instead a tuple
        content = """
from conan import ConanFile

class OpenSSLConan(ConanFile):
    name = "openssl"
    version = "2.0.1"
    exports = ['*.txt', '*.h']
"""
        self.client.save({CONANFILE: content})
        self.client.run("export . --user=lasote --channel=stable")
        layout = self.client.exported_layout()
        export_path = layout.export()
        assert sorted(os.listdir(export_path)) == ['CMakeLists.txt', CONANFILE, CONAN_MANIFEST]


class TestExportMetadata:

    def test_revision_mode_hash(self):
        t = TestClient(light=True)
        t.save({'conanfile.py': GenConanfile().with_revision_mode("hash")})
        t.run(f"export . --name=name --version=version")
        layout = t.exported_layout()
        assert layout.reference.revision == "99c241d3d47396772cd1ac5a36307a29"

    def test_revision_mode_invalid(self):
        t = TestClient(light=True)
        t.save({'conanfile.py': GenConanfile().with_revision_mode("auto")})
        t.run("export . --name=name --version=version", assert_error=True)
        assert "ERROR: Revision mode should be one of 'hash' (default) or 'scm'" in t.out

    def test_export_no_params(self):
        client = TestClient(light=True)
        client.save({"conanfile.py": GenConanfile("lib", "1.0")})
        client.run('export .')
        assert "lib/1.0: Exported" in client.out

    def test_export_with_name_and_version(self):
        client = TestClient(light=True)
        client.save({"conanfile.py": GenConanfile()})

        client.run('export . --name=lib --version=1.0')
        assert "lib/1.0: Exported" in client.out

    def test_export_with_only_user_channel(self):
        """This should be the recommended way and only from Conan 2.0"""
        client = TestClient(light=True)
        client.save({"conanfile.py": GenConanfile("lib", "1.0")})

        client.run('export .  --version= --user=user --channel=channel')
        assert "lib/1.0@user/channel: Exported" in client.out

    def test_export_conflict_no_user_channel(self):
        client = TestClient(light=True)
        client.save({"conanfile.py": GenConanfile()})

        client.run('export . --name=pkg --version=0.1 --user=user --channel=channel')
        assert "pkg/0.1@user/channel: Exported" in client.out
        client.run('export . --name=pkg --version=0.1 --user=other --channel=stable')
        assert "pkg/0.1@other/stable: Exported" in client.out
        client.run('export . --name=pkg --version=0.1')
        assert "pkg/0.1: Exported" in client.out
        client.run('export . --name=pkg --version=0.1')
        assert "pkg/0.1: Exported" in client.out


@pytest.mark.skipif(platform.system() != "Linux", reason="Needs case-sensitive filesystem")
def test_export_casing():
    # https://github.com/conan-io/conan/issues/8583
    client = TestClient(light=True)
    conanfile = textwrap.dedent("""
        from conan import ConanFile
        class Pkg(ConanFile):
            exports = "file1", "FILE1"
            exports_sources = "test", "TEST"
        """)
    client.save({"conanfile.py": conanfile,
                 "test": "some lowercase",
                 "TEST": "some UPPERCASE",
                 "file1": "file1 lowercase",
                 "FILE1": "file1 UPPERCASE"
                 })
    assert client.load("test") == "some lowercase"
    assert client.load("TEST") == "some UPPERCASE"
    assert client.load("file1") == "file1 lowercase"
    assert client.load("FILE1") == "file1 UPPERCASE"
    client.run("export . --name=pkg --version=0.1")
    layout = client.exported_layout()
    export_src_folder = layout.export_sources()
    assert load(os.path.join(export_src_folder, "test")) == "some lowercase"
    assert load(os.path.join(export_src_folder, "TEST")) == "some UPPERCASE"
    exports_folder = layout.export()
    assert load(os.path.join(exports_folder, "file1")) == "file1 lowercase"
    assert load(os.path.join(exports_folder, "FILE1")) == "file1 UPPERCASE"


def test_export_invalid_refs():
    c = TestClient(light=True)
    c.save({"conanfile.py": GenConanfile()})
    c.run("export . --name=pkg% --version=0.1", assert_error=True)
    assert "ERROR: Invalid package name 'pkg%'" in c.out
    c.run("export . --name=pkg --version=0.1%", assert_error=True)
    assert "ERROR: Invalid package version '0.1%'" in c.out
    c.run("export . --name=pkg --version=0.1 --user=user%", assert_error=True)
    assert "ERROR: Invalid package user 'user%'" in c.out
    c.run("export . --name=pkg --version=0.1 --user=user --channel=channel%", assert_error=True)
    assert "ERROR: Invalid package channel 'channel%'" in c.out


def test_allow_temp_uppercase():
    c = TestClient(light=True)
    c.save({"conanfile.py": GenConanfile()})
    c.run("export . --name=Pkg --version=0.1", assert_error=True)
    assert "ERROR: Conan packages names 'Pkg/0.1' must be all lowercase" in c.out
    c.save_home({"global.conf": "core:allow_uppercase_pkg_names=True"})
    c.run("export . --name=Pkg --version=0.1")
    assert "WARN: Package name 'Pkg/0.1' has uppercase, " \
           "and has been allowed by temporary config." in c.out


def test_warn_special_chars_refs():
    c = TestClient(light=True)
    c.save({"conanfile.py": GenConanfile()})
    c.run("export . --name=pkg.name --version=0.1")
    assert "WARN: Name containing special chars is discouraged 'pkg.name'" in c.out
    c.run("export . --name=name --version=0.1 --user=user+some")
    assert "WARN: User containing special chars is discouraged 'user+some'" in c.out
    c.run("export . --name=pkg.name --version=0.1 --user=user --channel=channel+some")
    assert "WARN: Channel containing special chars is discouraged 'channel+some'" in c.out


def test_export_json():
    c = TestClient(light=True)
    c.save({"conanfile.py": GenConanfile()})
    c.run("export . --name=foo --version=0.1 --format json")
    info = json.loads(c.stdout)
    assert info["reference"] == "foo/0.1#4d670581ccb765839f2239cc8dff8fbd"
    assert len(info) == 1  # Only "reference" key yet



import os
import textwrap

from conan.test.assets.genconanfile import GenConanfile
from conan.test.utils.tools import TestClient


def test_transitive_py_requires_revisions():
    # https://github.com/conan-io/conan/issues/5529
    client = TestClient(light=True)
    python_req = textwrap.dedent("""
        from conan import ConanFile
        some_var = {}
        class PackageInfo(ConanFile):
            pass
        """)
    conanfile = textwrap.dedent("""
        from conan import ConanFile
        class PackageInfo(ConanFile):
            python_requires = "dep/0.1@user/channel"
        """)
    consumer = textwrap.dedent("""
        from conan import ConanFile
        class MyConanfileBase(ConanFile):
            python_requires = "pkg/0.1@user/channel"
            def generate(self):
                self.output.info("VAR={}!!!".format(self.python_requires["dep"].module.some_var))
        """)
    client.save({"dep/conanfile.py": python_req.format("42"),
                 "pkg/conanfile.py": conanfile,
                 "consumer/conanfile.py": consumer})

    client.run("export dep --name=dep --version=0.1 --user=user --channel=channel")
    client.run("export pkg --name=pkg --version=0.1 --user=user --channel=channel")
    client.run("lock create consumer/conanfile.py")

    client.save({"dep/conanfile.py": python_req.format("123")})
    client.run("export dep --name=dep --version=0.1 --user=user --channel=channel")

    client.run("install consumer/conanfile.py")
    assert "conanfile.py: VAR=42!!!" in client.out

    os.remove(os.path.join(client.current_folder, "consumer/conan.lock"))
    client.run("install consumer/conanfile.py")
    assert "conanfile.py: VAR=123!!!" in client.out


def test_transitive_matching_revisions():
    client = TestClient(light=True)
    dep = textwrap.dedent("""
       from conan import ConanFile
       some_var = {}
       class PackageInfo(ConanFile):
           pass
       """)
    tool = textwrap.dedent("""
        from conan import ConanFile
        class PackageInfo(ConanFile):
            python_requires = "dep/{}"
            def package_id(self):
                self.output.info("VAR={{}}!!!".format(self.python_requires["dep"].module.some_var))
        """)
    pkg = textwrap.dedent("""
        from conan import ConanFile
        class MyConanfileBase(ConanFile):
            python_requires = "{}/0.1"
            def package_id(self):
                self.output.info("VAR={{}}!!!".format(self.python_requires["dep"].module.some_var))
        """)
    client.save({"dep/conanfile.py": dep.format(42),
                 "toola/conanfile.py": tool.format("0.1"),
                 "toolb/conanfile.py": tool.format("0.2"),
                 "pkga/conanfile.py": pkg.format("toola"),
                 "pkgb/conanfile.py": pkg.format("toolb"),
                 "app/conanfile.py": GenConanfile().with_requires("pkga/0.1", "pkgb/0.1")})

    client.run("export dep --name=dep --version=0.1")
    client.run("export dep --name=dep --version=0.2")
    client.run("export toola --name=toola --version=0.1")
    client.run("export toolb --name=toolb --version=0.1")
    client.run("create pkga --name=pkga --version=0.1")
    client.run("create pkgb --name=pkgb --version=0.1")
    client.run("lock create app/conanfile.py --lockfile-out=app.lock")

    client.save({"dep/conanfile.py": dep.format(123)})
    client.run("export dep --name=dep --version=0.1")
    client.run("export dep --name=dep --version=0.2")

    client.run("install app/conanfile.py --lockfile=app.lock")
    assert "pkga/0.1: VAR=42!!!" in client.out
    assert "pkgb/0.1: VAR=42!!!" in client.out
    assert "VAR=123" not in client.out

    client.run("install app/conanfile.py")
    assert "pkga/0.1: VAR=123!!!" in client.out
    assert "pkgb/0.1: VAR=123!!!" in client.out
    assert "VAR=42" not in client.out



import textwrap

from conan.test.utils.tools import TestClient


class TestInit:
    def test_wrong_init(self):
        client = TestClient()
        conanfile = textwrap.dedent("""
            from conan import ConanFile
            class Lib(ConanFile):
                def init(self):
                    random_error
            """)

        client.save({"conanfile.py": conanfile})
        client.run("export .", assert_error=True)
        assert "Error in init() method, line 5" in client.out
        assert "name 'random_error' is not defined" in client.out

    def test_init(self):
        client = TestClient()
        conanfile = textwrap.dedent("""
            from conan import ConanFile
            from conan.tools.files import load
            import os
            import json
            class Lib(ConanFile):
                exports = "data.json"
                def init(self):
                    data = load(self, os.path.join(self.recipe_folder, "data.json"))
                    d = json.loads(data)
                    self.license = d["license"]
                    self.description = d["description"]
                def export(self):
                    self.output.info("description: %s" % self.description)
                    self.output.info("license: %s" % self.license)
                def build(self):
                    self.output.info("description: %s" % self.description)
                    self.output.info("license: %s" % self.license)
            """)
        data = '{"license": "MIT", "description": "MyDescription"}'
        client.save({"conanfile.py": conanfile,
                     "data.json": data})

        client.run("export . --name=pkg --version=version")
        assert "description: MyDescription" in client.out
        assert "license: MIT" in client.out
        client.run("create . --name=pkg --version=0.1 --user=user --channel=testing")
        assert "description: MyDescription" in client.out
        assert "license: MIT" in client.out
